# -*- coding: utf-8 -*-
"""
Unet code based on https://github.com/ternaus/kaggle_dstl_submission/tree/master/src

TODO: review code and cleanup due to introduction of model factory
"""

from __future__ import division

import tensorflow as tf

from keras.models import model_from_json

'''
img_rows = 112          # -> renamed to input_height
img_cols = 112          # -> renamed to input_width
num_channels = 16       # -> renamed to nb_channels
num_mask_channels = 1   # -> renamed to nb_classes
'''

smooth = 1e-12

def jaccard_coef(y_true, y_pred):
    intersection = tf.keras.backend.sum(y_true * y_pred, axis=[0, -1, -2])
    sum_ = tf.keras.backend.sum(y_true + y_pred, axis=[0, -1, -2])

    jac = (intersection + smooth) / (sum_ - intersection + smooth)

    return tf.keras.backend.mean(jac)

def jaccard_coef_int(y_true, y_pred):
    y_pred_pos = tf.keras.backend.round(tf.keras.backend.clip(y_pred, 0, 1))

    intersection = tf.keras.backend.sum(y_true * y_pred_pos, axis=[0, -1, -2])
    sum_ = tf.keras.backend.sum(y_true + y_pred_pos, axis=[0, -1, -2])

    jac = (intersection + smooth) / (sum_ - intersection + smooth)

    return tf.keras.backend.mean(jac)

def jaccard_coef_loss(y_true, y_pred):
    return -tf.keras.backend.log(jaccard_coef(y_true, y_pred)) + tf.keras.backend.binary_crossentropy(y_pred, y_true) 

def get_model(input_width=256, input_height=256, nb_channels=3, nb_classes=1,
              loss_mode='binary_crossentropy', learning_rate=1e-4,
              init_model_weights: bool = False, pretrained_weights_filepath: str = None):
    
    inputs = tf.keras.layers.Input((input_width, input_height, nb_channels))
    conv1 = tf.keras.layers.Conv2D(32, 3, activation='relu', padding='same', kernel_initializer='he_uniform')(inputs)
    conv1 = tf.keras.layers.BatchNormalization(axis=-1)(conv1)
    #conv1 = keras.layers.advanced_activations.ELU()(conv1)
    conv1 = tf.keras.layers.Conv2D(32, 3, activation='relu', padding='same', kernel_initializer='he_uniform')(conv1)
    conv1 = tf.keras.layers.BatchNormalization(axis=-1)(conv1)
    #conv1 = tf.keras.layers.advanced_activations.ELU()(conv1)
    pool1 = tf.keras.layers.MaxPooling2D(pool_size=(2, 2))(conv1)

    conv2 = tf.keras.layers.Conv2D(64, 3, activation='relu', padding='same', kernel_initializer='he_uniform')(pool1)
    conv2 = tf.keras.layers.BatchNormalization(axis=-1)(conv2)
    #conv2 = keras.layers.advanced_activations.ELU()(conv2)
    conv2 = tf.keras.layers.Conv2D(64, 3, activation='relu', padding='same', kernel_initializer='he_uniform')(conv2)
    conv2 = tf.keras.layers.BatchNormalization(axis=-1)(conv2)
    #conv2 = tf.keras.layers.advanced_activations.ELU()(conv2)
    pool2 = tf.keras.layers.MaxPooling2D(pool_size=(2, 2))(conv2)

    conv3 = tf.keras.layers.Conv2D(128, 3, activation='relu', padding='same', kernel_initializer='he_uniform')(pool2)
    conv3 = tf.keras.layers.BatchNormalization(axis=-1)(conv3)
    #conv3 = keras.layers.advanced_activations.ELU()(conv3)
    conv3 = tf.keras.layers.Conv2D(128, 3, activation='relu', padding='same', kernel_initializer='he_uniform')(conv3)
    conv3 = tf.keras.layers.BatchNormalization(axis=-1)(conv3)
    #conv3 = tf.keras.layers.advanced_activations.ELU()(conv3)
    pool3 = tf.keras.layers.MaxPooling2D(pool_size=(2, 2))(conv3)

    conv4 = tf.keras.layers.Conv2D(256, 3, activation='relu', padding='same', kernel_initializer='he_uniform')(pool3)
    conv4 = tf.keras.layers.BatchNormalization(axis=-1)(conv4)
    #conv4 = keras.layers.advanced_activations.ELU()(conv4)
    conv4 = tf.keras.layers.Conv2D(256, 3, activation='relu', padding='same', kernel_initializer='he_uniform')(conv4)
    conv4 = tf.keras.layers.BatchNormalization(axis=-1)(conv4)
    #conv4 = keras.layers.advanced_activations.ELU()(conv4)
    pool4 = tf.keras.layers.MaxPooling2D(pool_size=(2, 2))(conv4)

    conv5 = tf.keras.layers.Conv2D(512, 3, activation='relu', padding='same', kernel_initializer='he_uniform')(pool4)
    conv5 = tf.keras.layers.BatchNormalization(axis=-1)(conv5)
    #conv5 = keras.layers.advanced_activations.ELU()(conv5)
    conv5 = tf.keras.layers.Conv2D(512, 3, activation='relu', padding='same', kernel_initializer='he_uniform')(conv5)
    conv5 = tf.keras.layers.BatchNormalization(axis=-1)(conv5)
    #conv5 = keras.layers.advanced_activations.ELU()(conv5)

    up6 = tf.keras.layers.Concatenate(axis=3)([tf.keras.layers.UpSampling2D(size=(2, 2))(conv5), conv4])
    #up6 = merge([tf.keras.layers.UpSampling2D(size=(2, 2))(conv5), conv4], mode='concat', concat_axis=1)
    conv6 = tf.keras.layers.Conv2D(256, 3, activation='relu', padding='same', kernel_initializer='he_uniform')(up6)
    conv6 = tf.keras.layers.BatchNormalization(axis=-1)(conv6)
    #conv6 = keras.layers.advanced_activations.ELU()(conv6)
    conv6 = tf.keras.layers.Conv2D(256, 3, activation='relu', padding='same', kernel_initializer='he_uniform')(conv6)
    conv6 = tf.keras.layers.BatchNormalization(axis=-1)(conv6)
    #conv6 = keras.layers.advanced_activations.ELU()(conv6)

    up7 = tf.keras.layers.Concatenate(axis=3)([tf.keras.layers.UpSampling2D(size=(2, 2))(conv6), conv3])
    #up7 = tf.keras.layers.merge([tf.keras.layers.UpSampling2D(size=(2, 2))(conv6), conv3], mode='concat', concat_axis=1)
    conv7 = tf.keras.layers.Conv2D(128, 3, activation='relu', padding='same', kernel_initializer='he_uniform')(up7)
    conv7 = tf.keras.layers.BatchNormalization(axis=-1)(conv7)
    #conv7 = keras.layers.advanced_activations.ELU()(conv7)
    conv7 = tf.keras.layers.Conv2D(128, 3, activation='relu', padding='same', kernel_initializer='he_uniform')(conv7)
    conv7 = tf.keras.layers.BatchNormalization(axis=-1)(conv7)
    #conv7 = keras.layers.advanced_activations.ELU()(conv7)

    up8 = tf.keras.layers.Concatenate(axis=3)([tf.keras.layers.UpSampling2D(size=(2, 2))(conv7), conv2])
    #up8 = tf.keras.layers.merge([tf.keras.layers.UpSampling2D(size=(2, 2))(conv7), conv2], mode='concat', concat_axis=1)
    conv8 = tf.keras.layers.Conv2D(64, 3, activation='relu', padding='same', kernel_initializer='he_uniform')(up8)
    conv8 = tf.keras.layers.BatchNormalization(axis=-1)(conv8)
    #conv8 = keras.layers.advanced_activations.ELU()(conv8)
    conv8 = tf.keras.layers.Conv2D(64, 3, activation='relu', padding='same', kernel_initializer='he_uniform')(conv8)
    conv8 = tf.keras.layers.BatchNormalization(axis=-1)(conv8)
    #conv8 = keras.layers.advanced_activations.ELU()(conv8)

    up9 = tf.keras.layers.Concatenate(axis=3)([tf.keras.layers.UpSampling2D(size=(2, 2))(conv8), conv1])
    #up9 = tf.keras.layers.merge([tf.keras.layers.UpSampling2D(size=(2, 2))(conv8), conv1], mode='concat', concat_axis=1)
    conv9 = tf.keras.layers.Conv2D(32, 3, activation='relu', padding='same', kernel_initializer='he_uniform')(up9)
    conv9 = tf.keras.layers.BatchNormalization(axis=-1)(conv9)
    #conv9 = keras.layers.advanced_activations.ELU()(conv9)
    conv9 = tf.keras.layers.Conv2D(32, 3, activation='relu', padding='same', kernel_initializer='he_uniform')(conv9)
    #crop9 = tf.keras.layers.Cropping2D(cropping=((16, 16), (16, 16)))(conv9)
    conv9 = tf.keras.layers.BatchNormalization(axis=-1)(conv9)
    #conv9 = keras.layers.advanced_activations.ELU()(conv9)
    
    # TODO: crop9 vervangen door conv9 als batnormalisation terug aan
    conv10 = tf.keras.layers.Conv2D(nb_classes, 1, activation='sigmoid')(conv9)

    model = tf.keras.models.Model(inputs=inputs, outputs=conv10)
    
    if loss_mode == "jaccard_coef_loss":
        loss_func = jaccard_coef_loss
    elif loss_mode == "binary_crossentropy":
        loss_func = 'binary_crossentropy'
    else:
        raise Exception(f"Unknown loss function: {loss_mode}")

    # TODO: probably this needs to be removed because of the introduction of 
    # the model factory
    model.compile(optimizer=tf.keras.optimizers.Nadam(lr=learning_rate), loss=loss_func, metrics=['binary_crossentropy', jaccard_coef_int])

    return model

'''
def flip_axis(x, axis):
    x = np.asarray(x).swapaxes(axis, 0)
    x = x[::-1, ...]
    x = x.swapaxes(0, axis)
    return x

def form_batch(X, y, batch_size):
    X_batch = np.zeros((batch_size, num_channels, img_rows, img_cols))
    y_batch = np.zeros((batch_size, num_mask_channels, img_rows, img_cols))
    X_height = X.shape[2]
    X_width = X.shape[3]

    for i in range(batch_size):
        random_width = random.randint(0, X_width - img_cols - 1)
        random_height = random.randint(0, X_height - img_rows - 1)

        random_image = random.randint(0, X.shape[0] - 1)

        y_batch[i] = y[random_image, :, random_height: random_height + img_rows, random_width: random_width + img_cols]
        X_batch[i] = np.array(X[random_image, :, random_height: random_height + img_rows, random_width: random_width + img_cols])
    return X_batch, y_batch


class threadsafe_iter:
    """Takes an iterator/generator and makes it thread-safe by
    serializing call to the `next` method of given iterator/generator.
    """
    def __init__(self, it):
        self.it = it
        self.lock = threading.Lock()

    def __iter__(self):
        return self

    def next(self):
        with self.lock:
            return self.it.next()


def threadsafe_generator(f):
    """A decorator that takes a generator function and makes it thread-safe.
    """
    def g(*a, **kw):
        return threadsafe_iter(f(*a, **kw))
    return g


@threadsafe_generator
def batch_generator(X, y, batch_size, horizontal_flip=False, vertical_flip=False, swap_axis=False):
    while True:
        X_batch, y_batch = form_batch(X, y, batch_size)

        for i in range(X_batch.shape[0]):
            xb = X_batch[i]
            yb = y_batch[i]

            if horizontal_flip:
                if np.random.random() < 0.5:
                    xb = flip_axis(xb, 1)
                    yb = flip_axis(yb, 1)

            if vertical_flip:
                if np.random.random() < 0.5:
                    xb = flip_axis(xb, 2)
                    yb = flip_axis(yb, 2)

            if swap_axis:
                if np.random.random() < 0.5:
                    xb = xb.swapaxes(1, 2)
                    yb = yb.swapaxes(1, 2)

            X_batch[i] = xb
            y_batch[i] = yb

        yield X_batch, y_batch[:, :, 16:16 + img_rows - 32, 16:16 + img_cols - 32]


def save_model(model, cross):
    json_string = model.to_json()
    if not os.path.isdir('cache'):
        os.mkdir('cache')
    json_name = 'architecture_' + cross + '.json'
    weight_name = 'model_weights_' + cross + '.h5'
    open(os.path.join('cache', json_name), 'w').write(json_string)
    model.save_weights(os.path.join('cache', weight_name), overwrite=True)


def save_history(history, suffix):
    filename = 'history/history_' + suffix + '.csv'
    pd.DataFrame(history.history).to_csv(filename, index=False)


def read_model(cross=''):
    json_name = 'architecture_' + cross + '.json'
    weight_name = 'model_weights_' + cross + '.h5'
    model = tf.keras.models(open(os.path.join('../src/cache', json_name)).read())
    model.load_weights(os.path.join('../src/cache', weight_name))
    return model
'''

'''
if __name__ == '__main__':
    data_path = '../data'
    now = datetime.datetime.now()

    print('[{}] Creating and compiling model...'.format(str(datetime.datetime.now())))

    model = get_unet0()

    print('[{}] Reading train...'.format(str(datetime.datetime.now())))
    f = h5py.File(os.path.join(data_path, 'train_16.h5'), 'r')

    X_train = f['train']

    y_train = np.array(f['train_mask'])[:, 1]
    y_train = np.expand_dims(y_train, 1)
    print(y_train.shape)

    train_ids = np.array(f['train_ids'])

    batch_size = 128
    nb_epoch = 50

    history = History()
    callbacks = [
        history,
    ]

    suffix = 'structures_3_'
    model.compile(optimizer=Nadam(lr=1e-3), loss=jaccard_coef_loss, metrics=['binary_crossentropy', jaccard_coef_int])
    model.fit_generator(batch_generator(X_train, y_train, batch_size, horizontal_flip=True, vertical_flip=True, swap_axis=True),
                        nb_epoch=nb_epoch,
                        verbose=1,
                        samples_per_epoch=batch_size * 400,
                        callbacks=callbacks,
                        nb_worker=8
                        )

    save_model(model, "{batch}_{epoch}_{suffix}".format(batch=batch_size, epoch=nb_epoch, suffix=suffix))
    save_history(history, suffix)

    suffix = 'structures_4_'
    model.compile(optimizer=Nadam(lr=1e-4), loss=jaccard_coef_loss, metrics=['binary_crossentropy', jaccard_coef_int])
    model.fit_generator(
        batch_generator(X_train, y_train, batch_size, horizontal_flip=True, vertical_flip=True, swap_axis=True),
        nb_epoch=nb_epoch,
        verbose=1,
        samples_per_epoch=batch_size * 400,
        callbacks=callbacks,
        )

    save_model(model, "{batch}_{epoch}_{suffix}".format(batch=batch_size, epoch=nb_epoch, suffix=suffix))
    save_history(history, suffix)
    f.close()
'''



# -*- coding: utf-8 -*-
"""
Module with specific helper functions to manage the configuration of orthoseg.
"""

import configparser
import json
import logging
from pathlib import Path
import tempfile
from typing import List, Optional

#-------------------------------------------------------------
# First define/init some general variables/constants
#-------------------------------------------------------------
# Get a logger...
logger = logging.getLogger(__name__)
#logger.setLevel(logging.DEBUG)

# Define the chars that cannot be used in codes that are use in filenames.
# Remark: '_' cannot be used because '_' is used as devider to parse filenames, and if it is 
# used in codes as well the parsing becomes a lot more difficult. 
illegal_chars_in_codes = ['_', ',', '.', '?', ':']

#-------------------------------------------------------------
# The real work
#-------------------------------------------------------------

def get_config_files(config_path: Path) -> List[Path]:

    ##### Init #####
    # First check input param
    config_path = config_path.expanduser()
    if not config_path.exists():
        raise Exception(f"Config file specified does not exist: {config_path}")

    # Collect the config files to use. The "hardcoded" defaults should always 
    # be loaded.
    script_dir = Path(__file__).resolve().parent.parent
    config_filepaths = [script_dir / 'project_defaults.ini']

    # Load possible extra config files from the project config file
    basic_projectconfig = configparser.ConfigParser(
            interpolation=configparser.ExtendedInterpolation(),
            converters={'list': lambda x: [i.strip() for i in x.split(',')]},
            allow_no_value=True)
    basic_projectconfig.read(config_path)

    default_basedir = config_path.parent
    extra_config_files_to_load = basic_projectconfig['general'].getlist('extra_config_files_to_load')
    if extra_config_files_to_load is not None:
        for config_file in extra_config_files_to_load:
            config_file_formatted = Path(config_file.format(
                    task_filepath=config_path, 
                    jobs_dir=config_path.parent)).expanduser()
            if not config_file_formatted.is_absolute():
                config_file_formatted = (default_basedir / config_file_formatted).resolve()
            config_filepaths.append(Path(config_file_formatted))

    # Finally add the specific project config file...
    config_filepaths.append(config_path)

    return config_filepaths

def read_config_ext(config_paths: List[Path]) -> configparser.ConfigParser:

    # Log config filepaths that don't exist...
    for config_filepath in config_paths:
        if not config_filepath.exists():
            logger.warning(f"config_filepath does not exist: {config_filepath}")

    ##### Now we are ready to read the entire configuration #####
    def parse_boolean_ext(input) -> Optional[bool]:
        if input is None:
            return None

        if input in ("True", "true", "1", 1):
            return True
        elif input in ("False", "false", "0", 0):
            return False
        elif input in ("False", "false", "0", 0):
            return None

    def safe_math_eval(string):
        """
        Function to evaluate a mathematical expression safely.
        """
        if string is None:
            return None
            
        allowed_chars = "0123456789+-*(). /"
        for char in string:
            if char not in allowed_chars:
                raise Exception("Unsafe eval")

        return eval(string)

    def to_path(pathlike: str) -> Optional[Path]:
        if pathlike is None: 
            return None
        else:
            if '{tempdir}' in pathlike:
                return Path(pathlike.format(tempdir=tempfile.gettempdir()))
            else:
                return Path(pathlike)

    config = configparser.ConfigParser(
            interpolation=configparser.ExtendedInterpolation(),
            converters={'list': lambda x: [i.strip() for i in x.split(',')],
                        'listint': lambda x: [int(i.strip()) for i in x.split(',')],
                        'listfloat': lambda x: [float(i.strip()) for i in x.split(',')],
                        'dict': lambda x: None if x is None else json.loads(x),
                        'path': lambda x: to_path(x),
                        'eval': lambda x: safe_math_eval(x),
                        'boolean_ext': lambda x: parse_boolean_ext(x)},
            allow_no_value=True)

    config.read(config_paths)
    return config
    
def as_dict(config: configparser.ConfigParser):
    """
    Converts a ConfigParser object into a dictionary.

    The resulting dictionary has sections as keys which point to a dict of the
    sections options as key => value pairs.
    """
    the_dict = {}
    for section in config.sections():
        the_dict[section] = {}
        for key, val in config.items(section):
            the_dict[section][key] = val
    return the_dict

# If the script is ran directly...
if __name__ == '__main__':
    raise Exception("Not implemented")
    


# -*- coding: utf-8 -*-
"""
Module with specific helper functions to manage the logging of orthoseg.

TODO: maybe it is cleaner to replace most code here by a config dict?
"""

from email.message import EmailMessage
import logging
import smtplib

from orthoseg.helpers import config_helper as conf

#-------------------------------------------------------------
# First define/init general variables/constants
#-------------------------------------------------------------
# Get a logger...
logger = logging.getLogger(__name__)

#-------------------------------------------------------------
# The real work
#-------------------------------------------------------------

def sendmail(
        subject: str, 
        body: str = None,
        stop_on_error: bool = False):

    if conf is None:
        raise Exception('Config is not initialized')
        
    if not conf.email.getboolean('enabled', fallback=False):
        return
    mail_from = conf.email.get('from', None)
    mail_to = conf.email.get('to', None)
    mail_server = conf.email.get('server', None)
    mail_server_username = conf.email.get('username', None)
    mail_server_password = conf.email.get('password', None)

    # If one of the necessary parameters not provided, log subject
    if(mail_from is None
       or mail_to is None
       or mail_server is None):
        logger.warning(f"Mail global_config not provided to send email with subject: {subject}")
        return

    try:
        # Create message
        msg = EmailMessage()
        msg.add_header('from', mail_from)
        msg.add_header('to', mail_to)
        msg.add_header('subject', subject)
        if body is not None:
            msg.set_payload(body)

        # Send the email
        server = smtplib.SMTP(mail_server)
        if(mail_server_username is not None
           and mail_server_password is not None):
            server.login(mail_server_username, mail_server_password)
        server.send_message(msg)
        server.quit()
    except Exception as ex:
        if stop_on_error is False:
            logger.exception("Error sending email")
        else:
            raise Exception("Error sending email") from ex

# If the script is ran directly...
if __name__ == '__main__':
    raise Exception("Not implemented")
    


# -*- coding: utf-8 -*-
"""
Download the sample project.
"""

import argparse
from pathlib import Path
import sys

# Because orthoseg isn't installed as package + it is higher in dir hierarchy, add root to sys.path
sys.path.insert(0, str(Path(__file__).resolve().parent.parent))
from orthoseg.util import gdrive_util
from orthoseg.util import git_downloader

def main():
    
    ### Interprete arguments ###
    parser = argparse.ArgumentParser(add_help=False)

    parser.add_argument('dest_dir', help='The directory to download the orthoseg sample projects to. Use ~ for you home directory.')
    args = parser.parse_args()
    dest_dir = Path(args.dest_dir).expanduser() / "orthoseg"

    dest_dir_full = dest_dir / "sample_projects"
    if dest_dir_full.exists():
        raise Exception(f"Destination directory already exists: {dest_dir_full}")

    ### Download ###        
    print(f"Start download of sample projects to {str(dest_dir_full)}")
    git_downloader.download(
            repo_url="https://github.com/orthoseg/orthoseg/tree/master/sample_projects", 
            output_dir=dest_dir)
    print(f"Download finished")
    print(f"Start download of footballfields pretrained neural net")
    footballfields_model_dir = dest_dir_full / "footballfields/models"
    model_hdf5_path = footballfields_model_dir / "footballfields_01_0.92512_242.hdf5"
    if model_hdf5_path.exists() is False:
        gdrive_util.download_file("1XmAenCW6K_RVwqC6xbkapJ5ws-f7-QgH", model_hdf5_path)
    model_hyperparams_path = footballfields_model_dir / "footballfields_01_hyperparams.json"
    if model_hyperparams_path.exists() is False:
        gdrive_util.download_file("1umxcd4RkB81sem9PdIpLoWeiIW8ga1u7", model_hyperparams_path)
    model_modeljson_path = footballfields_model_dir / "footballfields_01_model.json"
    if model_modeljson_path.exists() is False:
        gdrive_util.download_file("16qe8thBTrO3dFfLMU1T22gWcfHVXt8zQ", model_modeljson_path)
    print(f"Download finished")

if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
"""
Module with specific helper functions to manage the logging of orthoseg.

TODO: maybe it is cleaner to replace most code here by a config dict?
"""

import json
import logging
import logging.config
import datetime
from pathlib import Path
import tempfile

#-------------------------------------------------------------
# Log helpers
#-------------------------------------------------------------

class LoggingContext:
    def __init__(self, logger, level=None, handler=None, close=True):
        self.logger = logger
        self.level = level
        self.handler = handler
        self.close = close

    def __enter__(self):
        if self.level is not None:
            self.old_level = self.logger.level
            self.logger.setLevel(self.level)
        if self.handler:
            self.logger.addHandler(self.handler)

    def __exit__(self, et, ev, tb):
        if self.level is not None:
            self.logger.setLevel(self.old_level)
        if self.handler:
            self.logger.removeHandler(self.handler)
        if self.handler and self.close:
            self.handler.close()
        # implicit return of None => don't swallow exceptions

#-------------------------------------------------------------
# Init logging
#-------------------------------------------------------------

def init_logging_dictConfig(
        logconfig_path: Path = None, 
        logconfig_dict: dict = None,
        log_basedir: Path = None,
        loggername: str = None) -> logging.Logger:
    """
    Initializes the logging based on input in dictConfig format. 
    
    The input can be a dict or a json file.

    The added value of this function is:
        - it will format {iso_datetime} placeholders in handlers.*.filename
        - it will resolve relative file paths in handlers.*.filename to tempdir.

    Args:
        logconfig_path (Path, optional): Json file containing the dictConfig info. Defaults to None.
        logconfig_dict (dict, optional): Dict containing the dictConfig info. Defaults to None.
    """   
    # Init logging
    if logconfig_dict is None:
        if logconfig_path is not None and logconfig_path.exists():
            # Load from json file
            with open(logconfig_path, 'r') as logconfig_file:
                logconfig_dict = json.load(logconfig_file)
        else:
             raise Exception(f"If the logconfig_dict is None, the logconfig_path should point to an existing file: {logconfig_path}")

    # If there are file handlers, replace possible placeholders + make sure log dir exists
    assert logconfig_dict is not None
    for handler in logconfig_dict['handlers']:
        if "filename" in logconfig_dict['handlers'][handler]:
            # Format the filename
            log_path = Path(logconfig_dict['handlers'][handler]['filename'].format(
                    iso_datetime=f"{datetime.datetime.now():%Y-%m-%d_%H-%M-%S}"))

            # If the log_path is a relative path, resolve it to temp dir.
            if not log_path.is_absolute():
                if log_basedir is not None:
                    log_path = log_basedir / log_path
                else:
                    log_path = Path(tempfile.gettempdir()) / log_path
                print(f"Parameter logconfig.handlers.{handler}.filename was relative, so is now resolved to {log_path}")

            logconfig_dict['handlers'][handler]['filename'] = str(log_path)

            # Also make sure the log dir exists
            log_path.parent.mkdir(parents=True, exist_ok=True)

    # Now load the log config
    logging.config.dictConfig(logconfig_dict) # type: ignore

    return logging.getLogger(loggername)

def main_log_init(
        log_dir: Path,
        log_basefilename: str):

    # Check input parameters
    if not log_dir:
        raise Exception(f"Error: log_dir is mandatory!")
    
    # Make sure the log dir exists
    if not log_dir.exists():
        log_dir.mkdir(parents=True, exist_ok=True)
        
    # Get root logger
    logger = logging.getLogger('')
    
    # Set the general maximum log level...
    logger.setLevel(logging.INFO)
    for handler in logger.handlers:
        handler.flush()
        handler.close()
    
    # Remove all handlers and add the ones I want again, so a new log file is created for each run
    # Remark: the function removehandler doesn't seem to work?
    logger.handlers = []
    
    ch = logging.StreamHandler()
    ch.setLevel(logging.INFO)
    #ch.setFormatter(logging.Formatter('%(levelname)s|%(name)s|%(message)s'))
    #ch.setFormatter(logging.Formatter('%(asctime)s|%(levelname)s|%(name)s|%(message)s',
    #                                  datefmt='%H:%M:%S,uuu'))
    ch.setFormatter(logging.Formatter(fmt='%(asctime)s.%(msecs)03d|%(levelname)s|%(name)s|%(message)s',
                                      datefmt='%H:%M:%S')) 
    logger.addHandler(ch)
    
    log_filepath = log_dir / f"{datetime.datetime.now():%Y-%m-%d_%H-%M-%S}_{log_basefilename}.log"
    fh = logging.FileHandler(filename=str(log_filepath))
    fh.setLevel(logging.INFO)
    fh.setFormatter(logging.Formatter('%(asctime)s|%(levelname)s|%(name)s|%(message)s'))
    logger.addHandler(fh)
    
    return logger

def clean_log_dir(
        log_dir: Path,
        nb_logfiles_tokeep: int,
        pattern: str = '*.*'):
    """
    Clean a log dir.

    Args:
        log_dir (Path): dir with log files to clean.
        nb_logfiles_tokeep (int): the number of log files to keep. 
        pattern (str, optional): pattern of the file names of the log files. 
            Defaults to '*.*'.
    """
    # Check input params
    if log_dir is None or log_dir.exists() is False or nb_logfiles_tokeep is None:
        return
    
    # List log files and remove the ones that are too much 
    files = sorted(list(log_dir.glob(pattern)), reverse=True)
    if len(files) > nb_logfiles_tokeep:
        for file_index in range(nb_logfiles_tokeep, len(files)):
            files[file_index].unlink()

# If the script is ran directly...
if __name__ == '__main__':
    raise Exception("Not implemented")
    


# -*- coding: utf-8 -*-
"""
Module with functions for post-processing prediction masks towards polygons.
"""

import argparse
import logging
from pathlib import Path
import shlex
import sys
import traceback

from geofileops import geofileops

# Because orthoseg isn't installed as package + it is higher in dir hierarchy, add root to sys.path
sys.path.insert(0, str(Path(__file__).resolve().parent.parent))
from orthoseg.helpers import config_helper as conf
from orthoseg.helpers import email_helper
from orthoseg.lib import postprocess_predictions as postp
import orthoseg.model.model_helper as mh
from orthoseg.util import log_util

#-------------------------------------------------------------
# First define/init general variables/constants
#-------------------------------------------------------------
# Get a logger...
logger = logging.getLogger(__name__)

#-------------------------------------------------------------
# The real work
#-------------------------------------------------------------

def postprocess_argstr(argstr):
    args = shlex.split(argstr)
    postprocess_args(args)

def postprocess_args(args):

    ##### Interprete arguments #####
    parser = argparse.ArgumentParser(add_help=False)

    # Required arguments
    required = parser.add_argument_group('Required arguments')
    required.add_argument('-c', '--config', type=str, required=True,
            help="The config file to use")
    
    # Optional arguments
    optional = parser.add_argument_group('Optional arguments')
    # Add back help 
    optional.add_argument('-h', '--help', action='help', default=argparse.SUPPRESS,
            help='Show this help message and exit')

    # Interprete arguments
    args = parser.parse_args(args)

    ##### Run! #####
    postprocess(config_path=Path(args.config))

def postprocess(config_path: Path):
    """
    Postprocess the output of a prediction for the config specified.

    Args:
        config_path (Path): Path to the config file.
    """

    ##### Init #####  
    # Load the config and save in a bunch of global variables zo it 
    # is accessible everywhere 
    conf.read_orthoseg_config(config_path)
    
    # Init logging
    log_util.clean_log_dir(
            log_dir=conf.dirs.getpath('log_dir'),
            nb_logfiles_tokeep=conf.logging.getint('nb_logfiles_tokeep'))     
    global logger
    logger = log_util.main_log_init(conf.dirs.getpath('log_dir'), __name__)

    # Log start + send email
    message = f"Start postprocess for config {config_path.stem}"
    logger.info(message)
    logger.debug(f"Config used: \n{conf.pformat_config()}")    
    email_helper.sendmail(message)
    
    try:    
    
        # Create base filename of model to use
        # TODO: is force data version the most logical, or rather implement 
        #       force weights file or ?
        traindata_id = None
        force_model_traindata_id = conf.train.getint('force_model_traindata_id')
        if force_model_traindata_id is not None and force_model_traindata_id > -1:
            traindata_id = force_model_traindata_id 
        
        # Get the best model that already exists for this train dataset
        trainparams_id = conf.train.getint('trainparams_id')
        best_model = mh.get_best_model(
                model_dir=conf.dirs.getpath('model_dir'), 
                segment_subject=conf.general['segment_subject'],
                traindata_id=traindata_id,
                trainparams_id=trainparams_id)
        if best_model is None:
            raise Exception(f"No best model found in {conf.dirs.getpath('model_dir')}")

        # Input file  the "most recent" prediction result dir for this subject 
        output_vector_dir = conf.dirs.getpath('output_vector_dir')
        output_vector_name = f"{best_model['basefilename']}_{best_model['epoch']}_{conf.predict['image_layer']}"
        output_vector_path = output_vector_dir / f"{output_vector_name}.gpkg"
                
        # Prepare some parameters for the postprocessing
        nb_parallel = conf.general.getint('nb_parallel')

        dissolve = conf.postprocess.getboolean('dissolve')
        dissolve_tiles_path = conf.postprocess.getpath('dissolve_tiles_path')
        simplify_algorithm = conf.postprocess.get('simplify_algorithm')
        if simplify_algorithm is not None:
            simplify_algorithm = geofileops.SimplifyAlgorithm[simplify_algorithm]
            simplify_tolerance = conf.postprocess.geteval('simplify_tolerance')
            simplify_lookahead = conf.postprocess.get('simplify_lookahead')
            if simplify_lookahead is not None:
                simplify_lookahead = int(simplify_lookahead)

        ##### Go! #####
        postp.postprocess_predictions(
                input_path=output_vector_path,
                output_path=output_vector_path,
                dissolve=dissolve,
                dissolve_tiles_path=dissolve_tiles_path,
                simplify_algorithm=simplify_algorithm,
                simplify_tolerance=simplify_tolerance,
                simplify_lookahead=simplify_lookahead,
                nb_parallel=nb_parallel,
                force=False)

        # Log and send mail
        message = f"Completed postprocess for config {config_path.stem}"
        logger.info(message)
        email_helper.sendmail(message)
    except Exception as ex:
        message = f"ERROR while running postprocess for task {config_path.stem}"
        logger.exception(message)
        email_helper.sendmail(subject=message, body=f"Exception: {ex}\n\n {traceback.format_exc()}")
        raise Exception(message) from ex

def main():
    try:
        postprocess_args(sys.argv[1:])
    except Exception as ex:
        logger.exception(f"Error: {ex}")
        raise

# If the script is ran directly...
if __name__ == '__main__':
    main()


# -*- coding: utf-8 -*-
"""
Module with helper functions to create models.

Offers a common interface, regardless of the underlying model implementation 
and contains extra metrics, callbacks,...

Many models are supported by using this segmentation model zoo:
https://github.com/qubvel/segmentation_models
"""

import json
import logging
from pathlib import Path
import tempfile
from typing import Any, List

import numpy as np
import tensorflow as tf
import keras.models
import segmentation_models as sm

#-------------------------------------------------------------
# First define/init some general variables/constants
#-------------------------------------------------------------
# Get a logger...
logger = logging.getLogger(__name__)
#logger.setLevel(logging.INFO)

# Force using the keras implementation in tensorflow, because the default 
# use of this implementation seems broken with tensorflow 2.5
sm.set_framework('tf.keras')

#-------------------------------------------------------------
# The real work
#-------------------------------------------------------------

'''
preprocessing_fn = get_preprocessing('resnet34')
x = preprocessing_fn(x)
'''

def get_model(
        architecture: str,
        input_width: int = None,
        input_height: int = None,
        nb_channels: int = 3,
        nb_classes: int = 1,
        activation: str = 'softmax',
        init_weights_with: str = 'imagenet',
        freeze: bool = False) -> keras.models.Model:
    """
    Get a model.
    
    Args:
        architecture (str): Architecture of the network to create
        input_width (int, optional): Width of the input images. Defaults to None.
        input_height (int, optional): Height of the input images. Defaults to None.
        nb_channels (int, optional): Nb of channels/bands of the input images. Defaults to 3.
        nb_classes (int, optional): Nb of classes to be segmented to. Defaults to 1.
        activation (Activation, optional): Activation function of last layer. Defaults to 'softmax'.
        init_weights_with (str, optional): Weights to init the network with. Defaults to 'imagenet'.
        freeze (bool, optional): Freeze the final layer weights during 
            training. It is usefull to use this option for the first few 
            epochs get a more robust network. Defaults to False. 
    
    Raises:
        Exception: [description]
        Exception: [description]
    
    Returns:
        [type]: [description]
    """
    # Check architecture
    segment_architecture_parts = architecture.split('+')
    if len(segment_architecture_parts) < 2:
        raise Exception(f"Unsupported architecture: {architecture}")
    encoder = segment_architecture_parts[0]
    decoder = segment_architecture_parts[1]

    if decoder.lower() == 'unet':
        # These two unet variants are implemented in a seperate module
        if encoder.lower() == 'standard':
            logger.warn(f"Architecture {architecture} not tested in a long time, so use at own risk")
            import orthoseg.model.model_unet_standard as m
            if init_weights_with is not None:
                init_weights = True
            else:
                init_weights = False
            return m.get_model(input_width=input_width, input_height=input_height,
                               nb_channels=nb_channels, nb_classes=nb_classes,
                               init_model_weights=init_weights)
        elif encoder.lower() == 'ternaus':
            logger.warn(f"Architecture {architecture} not tested in a long time, so use at own risk")
            import orthoseg.model.model_unet_ternaus as m
            if init_weights_with is not None:
                init_weights = True
            else:
                init_weights = False                
            return m.get_model(input_width=input_width, input_height=input_height,
                               nb_channels=nb_channels, nb_classes=nb_classes,
                               init_model_weights=init_weights)

        # Some other unet variants is implemented using the segmentation_models library
        from segmentation_models import Unet
        #from segmentation_models.backbones import get_preprocessing

        model = Unet(backbone_name=encoder.lower(),
                     input_shape=(input_width, input_height, nb_channels),
                     classes=nb_classes, activation=activation,
                     encoder_weights=init_weights_with, encoder_freeze=freeze)
        return model
    elif decoder.lower() == 'pspnet':
        from segmentation_models import PSPNet
        #from segmentation_models.backbones import get_preprocessing

        model = PSPNet(backbone_name=encoder.lower(),
                       input_shape=(input_width, input_height, nb_channels),
                       classes=nb_classes, activation=activation,
                       encoder_weights=init_weights_with, encoder_freeze=freeze)
        return model
    elif decoder.lower() == 'linknet':
        from segmentation_models import Linknet
        #from segmentation_models.backbones import get_preprocessing

        # First check if input size is compatible with linknet 
        if input_width is not None and input_height is not None:
            check_image_size(decoder, input_width, input_height)
            
        model = Linknet(backbone_name=encoder.lower(),
                        input_shape=(input_width, input_height, nb_channels),
                        classes=nb_classes, activation=activation,
                        encoder_weights=init_weights_with, encoder_freeze=freeze)
        return model
    else:
        raise Exception(f"Unknown decoder architecture: {decoder}")

def compile_model(
        model: keras.models.Model,
        optimizer: str,
        optimizer_params: dict,
        loss: str,
        metrics: List[str] = None,
        sample_weight_mode: str = None,
        class_weights: list = None) -> keras.models.Model:
    """
    Compile the model for training.
    
    Args:
        model (keras.models.Model): the keras model to compile.
        optimizer (str): the optimizer to use.
        optimizer_params (dict): parameters to use for optimizer.
        loss (str): the loss function to use. One of:
            * categorical_crossentropy
            * weighted_categorical_crossentropy: class_weights should be specified!

        metrics (List[Metric], optional): metrics to use. Defaults to None. Possible values:
            * 
        sample_weight_mode (str, optional): sample weight mode to use. Defaults to None.
        class_weights (list, optional): class weigths to use. Defaults to None.
    """

    # If no merics specified, use default ones
    metric_funcs: List[Any] = []
    if metrics is None:
        if loss in ['categorical_crossentropy', 'weighted_categorical_crossentropy']:
            metric_funcs.append('categorical_accuracy')
        elif loss == 'sparse_categorical_crossentropy':
            #metrics.append('sparse_categorical_accuracy')
            None
        elif loss == 'binary_crossentropy':
            metric_funcs.append('binary_accuracy')

        iou_score = sm.metrics.IOUScore()
        metric_funcs.append(iou_score)
        f1_score = sm.metrics.FScore()
        metric_funcs.append(f1_score)
        #metric_funcs.append(jaccard_coef_round)
        # metric_funcs=[jaccard_coef, jaccard_coef_flat,
        #          jaccard_coef_int, dice_coef, 'accuracy', 'binary_accuracy']

    else:
        raise Exception("Specifying metrics not yet implemented")

    # Check loss function
    if loss == 'bcedice':
        loss_func = dice_coef_loss_bce
    elif loss == 'dice_loss':
        loss_func = sm.losses.DiceLoss()
    elif loss == 'jaccard_loss':
        loss_func = sm.losses.JaccardLoss()
    elif loss == 'weighted_categorical_crossentropy':
        if class_weights is None: 
            raise Exception(f"With loss == {loss}, class_weights cannot be None!")
        loss_func = weighted_categorical_crossentropy(class_weights)
    else:
        loss_func = loss

    # Create optimizer
    if optimizer == 'adam':
        optimizer_func = tf.keras.optimizers.Adam(**optimizer_params)
    else: 
        raise Exception(f"Error creating optimizer: {optimizer}, with params {optimizer_params}")

    logger.info(f"Compile model with optimizer: optimizer, loss: {loss}, class_weights: {class_weights}")
    model.compile(optimizer=optimizer_func, loss=loss_func, metrics=metric_funcs,
            sample_weight_mode=sample_weight_mode)

    return model

def load_model(
        model_to_use_filepath: Path, 
        compile: bool = True) -> keras.models.Model:
    """
    Load an existing model from a file.

    If loading the architecture + model from the file doesn't work, tries 
    different things to load a model anyway:
    1) looks for a ..._model.json file to load the architecture from,  
       the weights are then loaded from the original file
    2) looks for a ..._hyperparams.json file to create the architecture from, 
       the weights are then loaded from the original file

    Args:
        model_to_use_filepath (Path): The file to load the model from.
        compile (bool, optional): True to get a compiled version back that is 
            ready to train. Defaults to True.

    Raises:
        Exception: [description]

    Returns:
        tf.keras.models.Model: The loaded model.
    """
    errors = []
    model = None
    
    # If it is a file with the complete model, try loading it entirely...
    if not model_to_use_filepath.stem.endswith('_weights'):
        iou_score = sm.metrics.IOUScore()
        f1_score = sm.metrics.FScore()

        try:
            model = tf.keras.models.load_model(
                    str(model_to_use_filepath),
                    custom_objects={'jaccard_coef': jaccard_coef,
                                    'jaccard_coef_flat': jaccard_coef_flat,
                                    'jaccard_coef_round': jaccard_coef_round,
                                    'dice_coef': dice_coef,
                                    'iou_score': iou_score,
                                    'f1_score': f1_score,
                                    'weighted_categorical_crossentropy': weighted_categorical_crossentropy},
                    compile=compile)
        except Exception as ex:
            errors.append(f"Error loading model+weights from {model_to_use_filepath}: {ex}")
    
    # If no model returned yet, try loading loading architecture and weights seperately
    if model is None:
        # Load the architecture from a model.json file
        model_basestem = f"{'_'.join(model_to_use_filepath.stem.split('_')[0:2])}"

        # Check if there is a specific model.json file for these weights
        model_json_filepath = model_to_use_filepath.parent / (model_to_use_filepath.stem.replace('_weights', '') + '.json')
        if not model_json_filepath.exists():
            # If not, check if there is a model.json file for the training session
            model_json_filepath = model_to_use_filepath.parent / f"{model_basestem}_model.json"
        if model_json_filepath.exists():
            with model_json_filepath.open('r') as src:
                model_json = src.read()
            try:
                model = tf.keras.models.model_from_json(model_json)
            except Exception as ex:
                errors.append(f"Error loading model architecture from {model_json_filepath}: {ex}")
        else:
            errors.append(f"No model.json file found to load model from: {model_json_filepath}")
        
        # If loading model.json not successfull, create model based on hyperparams.json
        if model is None:
            # Load the hyperparams file if available
            hyperparams_json_filepath = model_to_use_filepath.parent / f"{model_basestem}_hyperparams.json"
            if hyperparams_json_filepath.exists():
                with hyperparams_json_filepath.open('r') as src:
                    hyperparams = json.load(src)
                # Create the model we want to use
                try:
                    model = get_model(
                            architecture=hyperparams['architecture']['architecture'],
                            nb_channels=hyperparams['architecture']['nb_channels'], 
                            nb_classes=hyperparams['architecture']['nb_classes'], 
                            activation=hyperparams['architecture']['activation_function'])
                except Exception as ex:
                    errors.append(f"Exception trying get_model() with parameters from: {hyperparams_json_filepath}: {ex}")
            else:
                errors.append(f"No hyperparams.json file found to load model from: {hyperparams_json_filepath}")

        # Now load the weights
        if model is not None:
            try:
                model.load_weights(str(model_to_use_filepath))
            except Exception as ex:
                errors.append(f"Exception trying model.load_weights on: {model_to_use_filepath}: {ex}")

    # Check if a model got loaded...
    if model is not None:
        # Eager seems to be 50% slower, tested on tensorflow 2.5
        model.run_eagerly = False
    else:
        # If we still have not model... time to give up.
        errors_str = ""
        if len(errors) > 0:
            errors_str = (" The following errors occured while trying: \n    -> " + "\n    -> ".join(errors))
        raise Exception(f"Error loading model for {model_to_use_filepath}.{errors_str}")

    return model # type: ignore

def set_trainable(
        model,
        recompile: bool = True):
    sm.utils.set_trainable(model=model, recompile=recompile) # doesn't seem to work, so save and load model
        
def check_image_size(
        decoder: str,
        input_width: int, 
        input_height: int):
    if decoder.lower() == 'linknet':
        if((input_width and (input_width % 16) != 0) 
           or (input_height and (input_height % 16) != 0)):
            message = f"STOP: input_width ({input_width} and input_height ({input_height}) should be divisable by 16!"
            logger.error(message)
            raise Exception(message)
    else:
        logger.info(f"check_image_size is not implemented for this model: {decoder}")
        
#------------------------------------------
# Loss functions
#------------------------------------------

def weighted_categorical_crossentropy(weights):
    """ weighted_categorical_crossentropy

        Args:
            * weights<ktensor|nparray|list>: crossentropy weights
        Returns:
            * weighted categorical crossentropy function
    """
    if isinstance(weights,list) or isinstance(weights, np.ndarray):
        weights=tf.keras.backend.variable(weights)

    def loss(target,output,from_logits=False):
        if not from_logits:
            output /= tf.reduce_sum(output,
                                    len(output.get_shape()) - 1,
                                    True)
            _epsilon = tf.convert_to_tensor(tf.keras.backend.epsilon(), dtype=output.dtype.base_dtype)
            output = tf.clip_by_value(output, _epsilon, 1. - _epsilon) # type: ignore
            weighted_losses = target * tf.math.log(output) * weights
            return - tf.reduce_sum(weighted_losses,len(output.get_shape()) - 1) # type: ignore
        else:
            raise ValueError('WeightedCategoricalCrossentropy: not valid with logits')
    return loss

def dice_coef_loss(y_true, y_pred):
    return 1 - dice_coef(y_true, y_pred)

def bootstrapped_crossentropy(y_true, y_pred, bootstrap_type='hard', alpha=0.95):
    target_tensor = y_true
    prediction_tensor = y_pred
    _epsilon = tf.convert_to_tensor(tf.keras.backend.epsilon(), prediction_tensor.dtype.base_dtype)
    prediction_tensor = tf.clip_by_value(prediction_tensor, _epsilon, 1 - _epsilon) # type: ignore
    prediction_tensor = tf.keras.backend.log(prediction_tensor / (1 - prediction_tensor)) # type: ignore

    if bootstrap_type == 'soft':
        bootstrap_target_tensor = alpha * target_tensor + (1.0 - alpha) * tf.sigmoid(prediction_tensor)
    else:
        bootstrap_target_tensor = alpha * target_tensor + (1.0 - alpha) * tf.cast(
            tf.sigmoid(prediction_tensor) > 0.5, tf.float32)
    return tf.keras.backend.mean(tf.nn.sigmoid_cross_entropy_with_logits(
        labels=bootstrap_target_tensor, logits=prediction_tensor))

def dice_coef_loss_bce(y_true, y_pred):
    dice = 0.5
    bce = 0.5
    bootstrapping = 'hard'
    alpha = 1.
    return bootstrapped_crossentropy(y_true, y_pred, bootstrapping, alpha) * bce + dice_coef_loss(y_true, y_pred) * dice

#------------------------------------------
# Metrics functions
#------------------------------------------

SMOOTH_LOSS = 1e-12

def jaccard_coef(y_true, y_pred):
    intersection = tf.keras.backend.sum(y_true * y_pred, axis=[0, -1, -2])
    sum_ = tf.keras.backend.sum(y_true + y_pred, axis=[0, -1, -2])

    jac = (intersection + SMOOTH_LOSS) / (sum_ - intersection + SMOOTH_LOSS)

    return tf.keras.backend.mean(jac)

def jaccard_coef_round(y_true, y_pred):
    y_pred_pos = tf.keras.backend.round(tf.keras.backend.clip(y_pred, 0, 1))

    intersection = tf.keras.backend.sum(y_true * y_pred_pos, axis=[0, -1, -2])
    sum_ = tf.keras.backend.sum(y_true + y_pred_pos, axis=[0, -1, -2])
    jac = (intersection + SMOOTH_LOSS) / (sum_ - intersection + SMOOTH_LOSS)
    return tf.keras.backend.mean(jac)

def jaccard_coef_flat(y_true, y_pred):
    y_true_f = tf.keras.backend.flatten(y_true)
    y_pred_f = tf.keras.backend.flatten(y_pred)
    intersection = tf.keras.backend.sum(y_true_f * y_pred_f)
    return (intersection + SMOOTH_LOSS) / (tf.keras.backend.sum(y_true_f) + tf.keras.backend.sum(y_pred_f) - intersection + SMOOTH_LOSS)

def dice_coef(y_true, y_pred, smooth=1.0):
    y_true_f = tf.keras.backend.flatten(y_true)
    y_pred_f = tf.keras.backend.flatten(y_pred)
    intersection = tf.keras.backend.sum(y_true_f * y_pred_f)
    return (2. * intersection + smooth) / (tf.keras.backend.sum(y_true_f) + tf.keras.backend.sum(y_pred_f) + smooth)

def pct_wrong(y_true, y_pred):
    y_pred_pos = tf.keras.backend.round(tf.keras.backend.clip(y_pred, 0, 1))

    intersection = tf.keras.backend.sum(y_true * y_pred_pos, axis=[0, -1, -2])
    sum_ = tf.keras.backend.sum(y_true + y_pred_pos, axis=[0, -1, -2])
    jac = (intersection + SMOOTH_LOSS) / (sum_ - intersection + SMOOTH_LOSS)
    return tf.keras.backend.mean(jac)



# -*- coding: utf-8 -*-
"""
Module to make it easy to start a training session.
"""

import argparse
import gc
import logging
import os
from pathlib import Path
import re
import shlex
import sys
import traceback
from typing import List, Optional

from tensorflow import keras as kr

# Because orthoseg isn't installed as package + it is higher in dir hierarchy, add root to sys.path
sys.path.insert(0, str(Path(__file__).resolve().parent.parent))
from orthoseg.helpers import config_helper as conf
from orthoseg.helpers import email_helper
from orthoseg.lib import prepare_traindatasets as prep
from orthoseg.lib import predicter
from orthoseg.lib import trainer
from orthoseg.model import model_factory as mf
from orthoseg.model import model_helper as mh
from orthoseg.util import log_util

#-------------------------------------------------------------
# First define/init general variables/constants
#-------------------------------------------------------------
# Get a logger...
logger = logging.getLogger(__name__)

#-------------------------------------------------------------
# The real work
#-------------------------------------------------------------

def _train_argstr(argstr):
    args = shlex.split(argstr)
    _train_args(args)

def _train_args(args):

    ##### Interprete arguments #####
    parser = argparse.ArgumentParser(add_help=False)

    # Required arguments
    required = parser.add_argument_group('Required arguments')
    required.add_argument('-c', '--config', type=str, required=True,
            help="The config file to use")
    
    # Optional arguments
    optional = parser.add_argument_group('Optional arguments')
    # Add back help 
    optional.add_argument('-h', '--help', action='help', default=argparse.SUPPRESS,
            help='Show this help message and exit')

    # Interprete arguments
    args = parser.parse_args(args)

    ##### Run! #####
    train(config_path=Path(args.config))

def train(config_path: Path):
    """
    Run a training session for the config specified.

    Args:
        config_path (Path): Path to the config file to use.
    """
    ##### Init #####  
    # Load the config and save in a bunch of global variables zo it 
    # is accessible everywhere 
    conf.read_orthoseg_config(config_path)
    
    # Init logging
    log_util.clean_log_dir(
            log_dir=conf.dirs.getpath('log_dir'),
            nb_logfiles_tokeep=conf.logging.getint('nb_logfiles_tokeep'))     
    global logger
    logger = log_util.main_log_init(conf.dirs.getpath('log_dir'), __name__)      
    
    # Log start + send email
    message = f"Start train for config {config_path.stem}"
    logger.info(message)
    logger.debug(f"Config used: \n{conf.pformat_config()}")    
    email_helper.sendmail(message)
    
    try:    

        # First check if the segment_subject has a valid name
        segment_subject = conf.general['segment_subject']
        if segment_subject == 'MUST_OVERRIDE':
            raise Exception("The segment_subject parameter needs to be overridden in the subject specific config file!!!") 
        elif '_' in segment_subject:
            raise Exception(f"The segment_subject parameter should not contain '_', so this is invalid: {segment_subject}!!!") 

        # Create the output dir's if they don't exist yet...
        for dir in [conf.dirs.getpath('project_dir'), conf.dirs.getpath('training_dir')]:
            if dir and not dir.exists():
                dir.mkdir()
        
        ##### If the training data doesn't exist yet, create it #####
        # Get the label input info
        label_files_dict = conf.train.getdict('label_datasources', None)
        label_infos = []
        if label_files_dict is not None:    
            for label_file_key in label_files_dict:
                label_file = label_files_dict[label_file_key]
                # Add as LabelInfo objects to list
                label_infos.append(prep.LabelInfo(
                        locations_path=Path(label_file['locations_path']),
                        polygons_path=Path(label_file['data_path']),
                        image_layer=label_file['image_layer']))
            if label_infos is None or len(label_infos) == 0:
                raise Exception(f"Parameter label_datasources is defined in config but doesn't contain valid label info!")
        else:
            # Search for the files based on the file name patterns... 
            labelpolygons_pattern = conf.train.getpath('labelpolygons_pattern')
            labellocations_pattern = conf.train.getpath('labellocations_pattern')
            label_infos = _search_label_files(labelpolygons_pattern, labellocations_pattern)
            if label_infos is None or len(label_infos) == 0:
                raise Exception(f"No label files found with patterns {labellocations_pattern} and {labelpolygons_pattern}")

        # Determine the projection of (the first) train layer... it will be used for all!!!
        train_image_layer = label_infos[0].image_layer
        train_projection = conf.image_layers[train_image_layer]['projection']

        # Determine classes 
        try:
            classes = conf.train.getdict('classes')

            # If the burn_value property isn't supplied for the classes, add them
            for class_id, (classname) in enumerate(classes):
                if 'burn_value' not in classes[classname]:
                    classes[classname]['burn_value'] = class_id
        except Exception as ex:
            raise Exception(f"Error reading classes: {conf.train.get('classes')}") from ex

        # Now create the train datasets (train, validation, test)
        force_model_traindata_id = conf.train.getint('force_model_traindata_id')
        if force_model_traindata_id > -1:
            training_dir = conf.dirs.getpath('training_dir') / f"{force_model_traindata_id:02d}"
            traindata_id = force_model_traindata_id
        else:
            logger.info("Prepare train, validation and test data")
            training_dir, traindata_id = prep.prepare_traindatasets(
                    label_infos=label_infos,
                    classes=classes,
                    image_layers=conf.image_layers,
                    training_dir=conf.dirs.getpath('training_dir'),
                    labelname_column=conf.train.get('labelname_column'),
                    image_pixel_x_size=conf.train.getfloat('image_pixel_x_size'),
                    image_pixel_y_size=conf.train.getfloat('image_pixel_y_size'),
                    image_pixel_width=conf.train.getint('image_pixel_width'),
                    image_pixel_height=conf.train.getint('image_pixel_height'),
                    ssl_verify=conf.general['ssl_verify'])
                    
        logger.info(f"Traindata dir to use is {training_dir}, with traindata_id: {traindata_id}")
        traindata_dir = training_dir / 'train'
        validationdata_dir = training_dir / 'validation'
        testdata_dir = training_dir / 'test'
        
        ##### Check if training is needed #####
        # Get hyper parameters from the config
        # TODO: activation_function should probably not be specified!!!!!!
        architectureparams = mh.ArchitectureParams(
                architecture=conf.model['architecture'],
                classes=[classname for classname in classes],
                nb_channels=conf.model.getint('nb_channels'),
                architecture_id=conf.model.getint('architecture_id'),
                activation_function='softmax')
        trainparams = mh.TrainParams(
                trainparams_id=conf.train.getint('trainparams_id'),
                image_augmentations=conf.train.getdict('image_augmentations'),
                mask_augmentations=conf.train.getdict('mask_augmentations'),
                class_weights=[classes[classname]['weight'] for classname in classes],
                batch_size=conf.train.getint('batch_size_fit'), 
                optimizer=conf.train.get('optimizer'), 
                optimizer_params=conf.train.getdict('optimizer_params'), 
                loss_function=conf.train.get('loss_function'), 
                monitor_metric=conf.train.get('monitor_metric'), 
                monitor_metric_mode=conf.train.get('monitor_metric_mode'), 
                save_format=conf.train.get('save_format'), 
                save_best_only=conf.train.getboolean('save_best_only'), 
                save_min_accuracy=conf.train.getfloat('save_min_accuracy'),
                nb_epoch=conf.train.getint('max_epoch'),
                nb_epoch_with_freeze=conf.train.getint('nb_epoch_with_freeze'),
                earlystop_patience=conf.train.getint('earlystop_patience'),
                earlystop_monitor_metric=conf.train.get('earlystop_monitor_metric'),
                earlystop_monitor_metric_mode=conf.train.get('earlystop_monitor_metric_mode'),
                log_tensorboard=conf.train.getboolean('log_tensorboard'),
                log_csv=conf.train.getboolean('log_csv'))

        # Check if there exists already a model for this train dataset + hyperparameters
        model_dir = conf.dirs.getpath('model_dir')
        segment_subject = conf.general['segment_subject']
        best_model_curr_train_version = mh.get_best_model(
                model_dir=model_dir, 
                segment_subject=segment_subject,
                traindata_id=traindata_id,
                architecture_id=architectureparams.architecture_id,
                trainparams_id=trainparams.trainparams_id)

        # Determine if training is needed,...
        resume_train = conf.train.getboolean('resume_train')
        if resume_train is False:
            # If no (best) model found, training needed!
            if best_model_curr_train_version is None:
                train_needed = True
            elif conf.train.getboolean('force_train') is True:
                train_needed = True
            else:
                logger.info("JUST PREDICT, without training: resume_train is false and model found")
                train_needed = False
        else:
            # We want to preload an existing model and models were found
            if best_model_curr_train_version is not None:
                logger.info(f"PRELOAD model and continue TRAINING it: {best_model_curr_train_version['filename']}")
                train_needed = True
            else:
                message = "STOP: preload_existing_model is true but no model was found!"
                logger.error(message)
                raise Exception(message)
        
        ##### Train!!! #####
        min_probability = conf.predict.getfloat("min_probability")
        if train_needed is True:

            # If a model already exists, use it to predict (possibly new) training and 
            # validation dataset. This way it is possible to have a quick check on errors
            # in (new) added labels in the datasets.

            # Get the current best model that already exists for this subject
            best_recent_model = mh.get_best_model(model_dir=model_dir)
            if best_recent_model is not None:
                try:
                    # TODO: move the hyperparams filename formatting to get_models...
                    logger.info(f"Load model + weights from {best_recent_model['filepath']}")    
                    best_model = mf.load_model(best_recent_model['filepath'], compile=False)
                    best_hyperparams_path = best_recent_model['filepath'].parent / f"{best_recent_model['basefilename']}_hyperparams.json"
                    best_hyperparams = mh.HyperParams(path=best_hyperparams_path)           
                    logger.info("Loaded model, weights and params")

                    # Prepare output subdir to be used for predictions
                    predict_out_subdir, _ = os.path.splitext(best_recent_model['filename'])
                    
                    # Predict training dataset
                    predicter.predict_dir(
                            model=best_model,
                            input_image_dir=traindata_dir / 'image',
                            output_image_dir=traindata_dir / predict_out_subdir,
                            output_vector_path=None,
                            projection_if_missing=train_projection,
                            input_mask_dir=traindata_dir / 'mask',
                            batch_size=conf.train.getint('batch_size_predict'), 
                            evaluate_mode=True,
                            classes=best_hyperparams.architecture.classes,
                            min_probability=min_probability,
                            cancel_filepath=conf.files.getpath('cancel_filepath'))
                        
                    # Predict validation dataset
                    predicter.predict_dir(
                            model=best_model,
                            input_image_dir=validationdata_dir / 'image',
                            output_image_dir=validationdata_dir / predict_out_subdir,
                            output_vector_path=None,
                            projection_if_missing=train_projection,
                            input_mask_dir=validationdata_dir / 'mask',
                            batch_size=conf.train.getint('batch_size_predict'), 
                            evaluate_mode=True,
                            classes=best_hyperparams.architecture.classes,
                            min_probability=min_probability,
                            cancel_filepath=conf.files.getpath('cancel_filepath'))
                    del best_model
                except Exception as ex:
                    logger.warn(f"Exception trying to predict with old model: {ex}")
            
            # Now we can really start training
            logger.info('Start training')
            model_preload_filepath = None
            if best_model_curr_train_version is not None:
                model_preload_filepath = best_model_curr_train_version['filepath']
            elif conf.train.getboolean('preload_with_previous_traindata'):
                best_model_for_architecture = mh.get_best_model(
                        model_dir=model_dir, 
                        segment_subject=segment_subject)
                if best_model_for_architecture is not None:
                    model_preload_filepath = best_model_for_architecture['filepath']
            
            # Combine all hyperparameters in hyperparams object
            hyperparams = mh.HyperParams(
                    architecture=architectureparams,
                    train=trainparams)

            trainer.train(
                    traindata_dir=traindata_dir,
                    validationdata_dir=validationdata_dir,
                    model_save_dir=model_dir,
                    segment_subject=segment_subject,
                    traindata_id=traindata_id,
                    hyperparams=hyperparams,
                    model_preload_filepath=model_preload_filepath,
                    image_width=conf.train.getint('image_pixel_width'),
                    image_height=conf.train.getint('image_pixel_height'),
                    save_augmented_subdir=conf.train.get('save_augmented_subdir')) 
        
            # Now get the best model found during training
            best_model_curr_train_version = mh.get_best_model(
                    model_dir=model_dir, 
                    segment_subject=segment_subject,
                    traindata_id=traindata_id)

        # Assert to evade typing warnings
        assert best_model_curr_train_version is not None
        
        # Now predict on the train,... data  
        logger.info(f"PREDICT test data with best model: {best_model_curr_train_version['filename']}")
        
        # Load prediction model...
        logger.info(f"Load model + weights from {best_model_curr_train_version['filepath']}")    
        model = mf.load_model(best_model_curr_train_version['filepath'], compile=False)            
        logger.info("Loaded model + weights")
        
        # Prepare output subdir to be used for predictions
        predict_out_subdir, _ = os.path.splitext(best_model_curr_train_version['filename'])
        
        # Predict training dataset
        predicter.predict_dir(
                model=model,
                input_image_dir=traindata_dir / 'image',
                output_image_dir=traindata_dir / predict_out_subdir,
                output_vector_path=None,
                projection_if_missing=train_projection,
                input_mask_dir=traindata_dir / 'mask',
                batch_size=conf.train.getint('batch_size_predict'), 
                evaluate_mode=True,
                classes=classes,
                min_probability=min_probability,
                cancel_filepath=conf.files.getpath('cancel_filepath'))
        
        # Predict validation dataset
        predicter.predict_dir(
                model=model,
                input_image_dir=validationdata_dir / 'image',
                output_image_dir=validationdata_dir / predict_out_subdir,
                output_vector_path=None,
                projection_if_missing=train_projection,
                input_mask_dir=validationdata_dir / 'mask',
                batch_size=conf.train.getint('batch_size_predict'), 
                evaluate_mode=True,
                classes=classes,
                min_probability=min_probability,
                cancel_filepath=conf.files.getpath('cancel_filepath'))

        # Predict test dataset, if it exists
        if testdata_dir is not None and testdata_dir.exists():
            predicter.predict_dir(
                    model=model,
                    input_image_dir=testdata_dir / 'image',
                    output_image_dir=testdata_dir / predict_out_subdir, 
                    output_vector_path=None,
                    projection_if_missing=train_projection,
                    input_mask_dir=testdata_dir / 'mask',
                    batch_size=conf.train.getint('batch_size_predict'), 
                    evaluate_mode=True,
                    classes=classes,
                    min_probability=min_probability,
                    cancel_filepath=conf.files.getpath('cancel_filepath'))
        
        # Predict extra test dataset with random images in the roi, to add to 
        # train and/or validation dataset if inaccuracies are found
        # -> this is very useful to find false positives to improve the datasets
        if conf.dirs.getpath('predictsample_image_input_dir').exists():
            predicter.predict_dir(
                    model=model,
                    input_image_dir=conf.dirs.getpath('predictsample_image_input_dir'),
                    output_image_dir=conf.dirs.getpath('predictsample_image_output_basedir') / predict_out_subdir,
                    output_vector_path=None,
                    projection_if_missing=train_projection,
                    batch_size=conf.train.getint('batch_size_predict'), 
                    evaluate_mode=True,
                    classes=classes,
                    min_probability=min_probability,
                    cancel_filepath=conf.files.getpath('cancel_filepath'))

        # Free resources...
        logger.debug("Free resources")
        if model is not None:
            del model
        kr.backend.clear_session()
        gc.collect()
    
        # Log and send mail
        message = f"Completed train for config {config_path.stem}"
        logger.info(message)
        email_helper.sendmail(message)
    except Exception as ex:
        message = f"ERROR while running train for task {config_path.stem}"
        logger.exception(message)
        email_helper.sendmail(subject=message, body=f"Exception: {ex}\n\n {traceback.format_exc()}")
        raise Exception(message) from ex

def _search_label_files(
        labelpolygons_pattern: Path,
        labellocations_pattern: Path) -> List[prep.LabelInfo]:
    label_infos = []

    labelpolygons_pattern_searchpath = Path(str(labelpolygons_pattern).format(image_layer='*'))
    labelpolygons_paths = list(labelpolygons_pattern_searchpath.parent.glob(labelpolygons_pattern_searchpath.name))
    labellocations_pattern_searchpath = Path(str(labellocations_pattern).format(image_layer='*'))
    labellocations_paths = list(labellocations_pattern_searchpath.parent.glob(labellocations_pattern_searchpath.name))

    # Loop through all labellocation files
    for labellocations_path in labellocations_paths:
        tokens = _unformat(labellocations_path.stem, labellocations_pattern.stem)
        if 'image_layer' not in tokens:
            raise Exception(f"image_layer token not found in {labellocations_path} using pattern {labellocations_pattern}")
        image_layer = tokens['image_layer']

        # Look for the matching (= same image_layer) data file
        found = False
        for labelpolygons_path in labelpolygons_paths:
            tokens = _unformat(labelpolygons_path.stem, labelpolygons_pattern.stem)
            if 'image_layer' not in tokens:
                raise Exception(f"image_layer token not found in {labelpolygons_path} using pattern {labelpolygons_pattern}")
            
            if tokens['image_layer'] == image_layer:
                found = True
                label_infos.append(prep.LabelInfo(
                        locations_path=labellocations_path,
                        polygons_path=labelpolygons_path,
                        image_layer=image_layer))
        if found is False:
            raise Exception(f"No matching data file found for {labellocations_path}")
    
    return label_infos

def _unformat(string: str, pattern: str) -> dict:
    regex = re.sub(r'{(.+?)}', r'(?P<_\1>.+)', pattern)
    regex_result = re.search(regex, string)
    if regex_result is not None:
        values = list(regex_result.groups())
        keys = re.findall(r'{(.+?)}', pattern)
        _dict = dict(zip(keys, values))
        return _dict
    else:
        raise Exception(f"Error: pattern {pattern} not found in {string}")

def main():
    try:
        _train_args(sys.argv[1:])
    except Exception as ex:
        logger.exception(f"Error: {ex}")
        raise
    
# If the script is ran directly...
if __name__ == '__main__':
    main()
    


# -*- coding: utf-8 -*-
"""
Module with helper functions regarding (keras) models.
"""

import json
import logging
from pathlib import Path
import shutil
from typing import List, Optional

import pandas as pd 
from keras import callbacks

#-------------------------------------------------------------
# First define/init some general variables/constants
#-------------------------------------------------------------
# Get a logger...
logger = logging.getLogger(__name__)
#logger.setLevel(logging.INFO)

#-------------------------------------------------------------
# The real work
#-------------------------------------------------------------
class ArchitectureParams:

    def __init__(
            self,
            architecture: str,
            classes: list = None,
            nb_channels: int = 3,
            activation_function: str = 'softmax',
            architecture_id: int = 0):
        """
        Class containing the hyper parameters needed to create the model.
        
        Args:
            architecture (str): the model architecture to use.
            classes (list, optional): list of classes that will be detected by 
                the model. Default to None, and then 2 generic classes are 
                supposed. 
            nb_channels (int, optional): Number of channels of the images. 
                Defaults to 3.
            activation_function (str, optional): activation function to use. 
                Defaults to softmax. 
            architecture_id (int, optional): id of the architecture. 
                Defaults to 0. 
        """
        self.architecture = architecture
        if classes is not None:
            self.classes = classes
        else:
            self.classes = ['background', 'segmentsubject']
        self.nb_channels = nb_channels
        self.activation_function = activation_function
        self.architecture_id = architecture_id

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__, sort_keys=True, indent=4)

class TrainParams:
    
    def __init__(
            self,
            image_augmentations: dict, 
            mask_augmentations: dict,                        
            trainparams_id: int = 0,
            class_weights: list = None,
            batch_size: int = 4,
            optimizer: str = 'adam',
            optimizer_params: dict = None,
            loss_function: str = None,
            monitor_metric: str = None,
            monitor_metric_mode: str = 'auto',
            save_format: str = 'h5',
            save_best_only: bool = True,
            save_min_accuracy: float = 0.95,
            nb_epoch: int = 1000,
            nb_epoch_with_freeze: int = 100,
            earlystop_patience: int = 100,
            earlystop_monitor_metric: str = None,
            earlystop_monitor_metric_mode: str = 'auto',
            log_tensorboard: bool = False,
            log_csv: bool = True):
        """
        Class containing the hyper parameters needed to perform a training.
        
        Args:
            image_augmentations (dict): The augmentations to use on the input image during training.
            mask_augmentations (dict): The augmentations to use on the input mask during training.
            architecture_id (int, optional): id of the architecture. Defaults to 0. 
            trainparams_id (int, optional): id of the hyperparams. Defaults to 0.
            class_weights (list, optional): [description]. Defaults to None.
            batch_size (int, optional): batch size to use while training. This must be 
                choosen depending on the neural network architecture
                and available memory on you GPU. Defaults to 4.
            optimizer (str, optional): Optimizer to use for training. Defaults to 'adam'.
            optimizer_params (dict, optional): Optimizer params to use. Defaults to { 'learning_rate': 0.0001 }.
            loss_function (str, optional): [description]. Defaults to None.
            monitor_metric (str, optional): Metric to monitor. If not specified 
                the loss function will drive the metric. Defaults to None.
            monitor_metric_mode (str, optional): Mode of the metric to monitor. 
                Defaults to 'auto'.
            save_format (str, optional): [description]. Defaults to 'h5'.
            save_best_only (bool, optional): [description]. Defaults to True.
            save_min_accuracy (float, optional): minimum accuracy to save a model. Defaults to 0.95.
            nb_epoch (int, optional): maximum number of epochs to train. Defaults to 1000.
            nb_epoch_with_freeze (int, optional): number epochs to train with 
                    part of the layers frozen. Defaults to 20.
            earlystop_patience (int, optional): [description]. Defaults to 100.
            earlystop_monitor_metric (str, optional): [description]. Defaults to None.
            earlystop_monitor_metric_mode (str, optional): Mode to monitor the 
                metric: 'max' if the metric should be as high as possible, 
                'min' if it should be low. Defaults to 'auto'.
            log_tensorboard (bool, optional): True to activate tensorboard 
                logging. Defaults to False.
            log_csv (bool, optional): True to activate logging to a csv. 
                Defaults to True
        
        Raises:
            Exception: [description]
        """
        self.trainparams_id = trainparams_id
        self.image_augmentations = image_augmentations
        self.mask_augmentations = mask_augmentations
        self.class_weights = class_weights
        self.batch_size = batch_size    
        
        self.optimizer = optimizer
        if optimizer_params is None:
            # Best set to 0.0001 to start (1e-3 is not ok)
            self.optimizer_params = { 'learning_rate': 0.0001 } 
        else:
            self.optimizer_params = optimizer_params

        if self.class_weights is not None: 
            self.loss_function = 'weighted_categorical_crossentropy'
        else:
            self.loss_function = 'categorical_crossentropy'

        # Properties to choose the best model
        if monitor_metric is not None:
            self.monitor_metric = monitor_metric
        elif self.loss_function in (
                'weighted_categorical_crossentropy', 'categorical_crossentropy'):
            self.monitor_metric = 'categorical_accuracy'
        self.monitor_metric_mode = monitor_metric_mode

        self.save_format = save_format
        self.save_best_only = save_best_only
        self.save_min_accuracy=save_min_accuracy
        self.nb_epoch = nb_epoch
        self.nb_epoch_with_freeze = nb_epoch_with_freeze
        
        # Properties to stop the training
        self.earlystop_patience = earlystop_patience
        if earlystop_monitor_metric is not None:
            self.earlystop_monitor_metric = earlystop_monitor_metric
        else:
            self.earlystop_monitor_metric = self.monitor_metric
        self.earlystop_monitor_metric_mode = earlystop_monitor_metric_mode

        # Properties regarding logging
        self.log_tensorboard = log_tensorboard 
        self.log_csv = log_csv

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__, sort_keys=True, indent=4)

class HyperParams:

    def __init__(
            self,
            architecture: ArchitectureParams = None,
            train: TrainParams = None,
            path: Path = None):
        """
        Class to store the hyper parameters to use for the machine learning algorythm.
            
        Args:
            architecture (ArchitectureParams): the fixed parameters that define the architecture of the 
                neural network. When training a network, it is possible to reuse the 
                weights of another network if these parameters are the same.
            train (TrainParams): these are the parameters that can be changed with each training.
        """
        self.fileversion = 1.1
        if architecture is not None:
            self.architecture=architecture
        if train is not None:
            self.train=train
        if path is not None:
            with open(path, 'r') as jsonfile:
                jsonstr = jsonfile.read()
                data = json.loads(jsonstr)

                # Read file version if it is present
                if 'fileversion' in data:
                    self.fileversion = data['fileversion']

                # Now parse the real data
                # For backwards compatibility, remove 'nb_classes' parameter if it exists
                if 'nb_classes' in data['architecture']:
                    del data['architecture']['nb_classes']

                self.architecture = ArchitectureParams(**data['architecture'])
                self.train = TrainParams(**data['train'])

    def toJSON(self):
        return json.dumps(self, default=lambda o: o.__dict__, sort_keys=True, indent=4)

def format_model_basefilename(
        segment_subject: str,
        traindata_id: int,
        architecture_id: int = 0,
        trainparams_id: int = 0) -> str:
    """
    Format the parameters into a model_filename.
    
    Args
        segment_subject: the segment subject
        traindata_id: the id of the data used to train the model
        architecture_id: the id of the model architecture used
        trainparams_id: the id of the hyper parameters used to train the model
    """
    # Format file name
    filename = f"{segment_subject}_{traindata_id:02d}"

    # If an architecture id and/or hyperparams id is specified, add those as well
    if architecture_id > 0 or trainparams_id > 0:
        filename = f"{filename}.{architecture_id}.{trainparams_id}"
    return filename

def format_model_filename(
        segment_subject: str,
        traindata_id: int,
        architecture_id: int,
        trainparams_id: int,
        acc_train: float,
        acc_val: float,
        acc_combined: float,
        epoch: int,
        save_format: str) -> str:
    """
    Format the parameters into a model_filename.
    
    Args
        segment_subject: the segment subject
        traindata_id: the version of the data used to train the model
        architecture_id: the id of the model architecture used
        trainparams_id: the version of the hyper parameters used to train
        acc_train: the accuracy reached for the training dataset
        acc_val: the accuracy reached for the validation dataset
        acc_combined: the average of the train and validation accuracy
        epoch: the epoch during training that reached these model weights
        save_format (str): the format to save in:
            * keras format: 'h5'
            * tensorflow savedmodel: 'tf'
    """
    # Format file name
    filename = format_model_basefilename(
            segment_subject=segment_subject,
            traindata_id=traindata_id,
            architecture_id=architecture_id,
            trainparams_id=trainparams_id)
    filename += f"_{acc_combined:.5f}_{epoch}"

    # Add suffix
    if save_format == 'tf':
        filename += "_tf"
    else:
        filename += ".hdf5"

    return filename

def parse_model_filename(filepath: Path) -> Optional[dict]:
    """
    Parse a model_filename to a dict containing the properties of the model:
        * segment_subject: the segment subject
        * traindata_id: the version of the data used to train the model
        * acc_combined: the average of the train and validation accuracy
        * trainparams_id: the version of the hyper parameters used to train
        * epoch: the epoch during training that reached these model weights
        * save_format (str): the format to save in:
            * keras format: 'h5'
            * tensorflow savedmodel: 'tf'
    
    Args
        filepath: the filepath to the model file
    """
    
    # Prepare filepath to extract info
    if filepath.is_dir():
        # If it is a dir, it should end on _tf
        if not filepath.name.endswith("_tf"):
            logger.warning(f"Not a valid path for a model, dir needs to end on _tf: {filepath}")
            return None
        save_format = 'tf'
        filename = filepath.name
    else:
        filename = filepath.stem
        if filepath.suffix in ('.h5', '.hdf5'):
            save_format = 'h5'
        else: 
            logger.warning(f"Not a valid path for a model, file should have .h5 of .hdf5 as suffix: {filepath}")
            return None
        
    # Now extract the fields...
    param_values = filename.split("_")
    if len(param_values) < 3:
        logger.warning(f"Not a valid path for a model, split('_') should result in >= 2 fields: {filepath}")

    segment_subject = param_values[0]

    model_info = param_values[1]
    model_info_values = model_info.split('.')   
    traindata_id = int(model_info_values[0])
    architecture_id = 0
    trainparams_id = 0
    if len(model_info_values) > 1:
        architecture_id = int(model_info_values[1])
        if len(model_info_values) > 2:
            trainparams_id = int(model_info_values[2])
    
    acc_combined = float(param_values[2])
    epoch = int(param_values[3])
        
    basefilename = format_model_basefilename(
            segment_subject=segment_subject,
            traindata_id=traindata_id,
            architecture_id=architecture_id,
            trainparams_id=trainparams_id)

    return {'filepath': filepath,
            'filename': filename,
            'basefilename': basefilename,
            'segment_subject': segment_subject,
            'traindata_id': traindata_id,
            'architecture_id': architecture_id,
            'trainparams_id': trainparams_id,
            'acc_combined': acc_combined,
            'epoch': epoch,
            'save_format': save_format}

def get_models(
        model_dir: Path,
        segment_subject: str = None,
        traindata_id: int = None,
        architecture_id: int = None,
        trainparams_id: int = None) -> List[dict]:
    """
    Return the list of models in the model_dir passed. It is returned as a 
    dataframe with the columns as returned in parse_model_filename()
    
    Args
        model_dir (Path): dir containing the models
        segment_subject (str, optional): only models with this the segment subject 
        traindata_id (int, optional): only models with this traindata version
        architecture_id (int, optional): only models with this this architecture_id
        trainparams_id (int, optional): only models with this hyperparams version
    """

    # List models
    model_paths = []
    model_paths.extend(model_dir.glob('*.hdf5'))
    model_paths.extend(model_dir.glob('*.h5'))
    model_paths.extend(model_dir.glob('*_tf'))

    # Loop through all models and extract necessary info...
    model_info_list = []
    for model_path in model_paths:
        model_info = parse_model_filename(model_path)
        if model_info is not None:
            model_info_list.append(model_info)

    # Filter, if filters provided
    if len(model_info_list) > 0:
        if segment_subject is not None:
            model_info_list = [model_info for model_info in model_info_list 
                               if model_info['segment_subject'] == segment_subject]
        if traindata_id is not None:
            model_info_list = [model_info for model_info in model_info_list 
                               if model_info['traindata_id'] == traindata_id]
        if trainparams_id is not None:
            model_info_list = [model_info for model_info in model_info_list 
                               if model_info['trainparams_id'] == trainparams_id]
        if architecture_id is not None:
            model_info_list = [model_info for model_info in model_info_list 
                               if model_info['architecture_id'] == architecture_id]

    return model_info_list

def get_best_model(
        model_dir: Path,
        segment_subject: str = None,
        traindata_id: int = None,
        architecture_id: int = None,
        trainparams_id: int = None) -> Optional[dict]:
    """
    Get the properties of the model with the highest combined accuracy for the highest 
    traindata version in the dir.

    Remark: regardless of the monitor function used when training, the accuracies
    are always better if higher!
    
    Args
        model_dir: dir containing the models
        segment_subject (str, optional): only models with this the segment subject 
        traindata_id (int, optional): only models with this train data id
        architecture_id (int, optional): only models with this the architecture_id
        trainparams_id (int, optional): only models with this hyperparams id
        
    Returns
        A dictionary with the info of the best model, or None if no model was found
    """
    # Get list of existing models for this train dataset
    model_info_list = get_models(
            model_dir=model_dir,
            segment_subject=segment_subject,
            traindata_id=traindata_id,
            architecture_id=architecture_id,
            trainparams_id=trainparams_id)
    
    # If nothing found, return None
    if len(model_info_list) == 0:
        return None

    # If no traindata_id provided, find highest traindata id
    max_traindata_id = -1
    for model_info in model_info_list:
        if model_info['traindata_id'] > max_traindata_id:
            max_traindata_id = model_info['traindata_id']
    
    # Get the list of newest model
    newest_model_info_list = [model_info for model_info in model_info_list 
                              if model_info['traindata_id'] == max_traindata_id]

    # If only one result, return it
    if len(newest_model_info_list) == 1:
        return newest_model_info_list[0]
    else:
        max_acc_combined = -1
        max_acc_combined_idx = -1
        for model_info_idx, model_info in enumerate(model_info_list):
            if model_info['acc_combined'] > max_acc_combined:
                max_acc_combined = model_info['acc_combined']
                max_acc_combined_idx = model_info_idx
    
        return model_info_list[max_acc_combined_idx]
    
class ModelCheckpointExt(callbacks.Callback):
    
    def __init__(
            self, 
            model_save_dir: Path,
            segment_subject: str,
            traindata_id: int,
            architecture_id: int,
            trainparams_id: int,
            monitor_metric: str,
            monitor_metric_mode: str,
            save_format: str = 'h5',
            save_best_only: bool = False,
            save_min_accuracy: float = 0.90,
            save_min_accuracy_ignored_epoch: int = 100,
            save_weights_only: bool = False,
            model_template_for_save = None,
            verbose: bool = True,
            only_report: bool = False):
        """
        Constructor
        
        Args:
            model_save_dir (Path): [description]
            segment_subject (str): segment subject 
            traindata_id (int): train data id
            architecture_id (int): model architecture id
            trainparams_id (int): id of the hyper parameters used
            monitor_metric (str): The metric to monitor for accuracy
            monitor_metric_mode (str): use 'min' if the accuracy metrics 
                should be  as low as possible, 'max' if a higher values is 
                better. 
            save_format (str, optional): The format to save in: 
                'h5' (keras format) or 'tf' (tensorflow savedmodel). 
                Defaults to 'tf'
            save_best_only (bool, optional): only keep the best model. 
                Defaults to True.
            save_min_accuracy (float, optional): minimum accuracy to be 
                reached to save model. Defaults to 0.9.
            save_min_accuracy_ignored_epoch (int, optional): at this epoch the 
                save_min_accuracy is ignored, to make sure there is a model  
                saved, even if the minimum accuracy is never reached. 
            save_weights_only: optional: only save weights
            model_template_for_save: optional, if using multi-GPU training, 
                pass the original model here to use this as template for saving       
            verbose (bool, optional): [description]. Defaults to True.
            only_report (bool, optional): [description]. Defaults to False.
        """
        monitor_metric_mode_values = ['min', 'max']
        if(monitor_metric_mode not in monitor_metric_mode_values):
            raise Exception(f"Invalid value for mode: {monitor_metric_mode}, should be one of {monitor_metric_mode_values}")
        save_format_values = ('h5', 'tf')
        if save_format not in save_format_values:
            raise Exception(f"Invalid value for save_format: {save_format}, should be one of {save_format_values}")
        
        self.model_save_dir = model_save_dir
        self.segment_subject = segment_subject
        self.traindata_id = traindata_id
        self.architecture_id = architecture_id
        self.trainparams_id = trainparams_id
        self.monitor_metric = monitor_metric
        self.monitor_metric_mode = monitor_metric_mode
        self.save_format = save_format
        self.save_best_only = save_best_only
        self.save_min_accuracy = save_min_accuracy
        self.save_min_accuracy_ignored_epoch = save_min_accuracy_ignored_epoch
        self.save_weights_only = save_weights_only
        self.model_template_for_save = model_template_for_save
        self.verbose = verbose
        self.only_report = only_report
        
    def on_epoch_end(self, epoch, logs={}):
        logger.debug(f"Start in callback on_epoch_begin, logs contains: {logs}")
        
        # First determine the values of the monitor metric for train and validation
        # If the monitor_metric doesn't contain placeholder, it is easy
        if '{' not in self.monitor_metric:
            new_model_monitor_train = logs.get(self.monitor_metric)
            new_model_monitor_val = logs.get('val_' + self.monitor_metric)
        else:
            # There are placeholders, so we need some more logic
            monitor_metric_train_formatted = self.monitor_metric.format(**logs)
            new_model_monitor_train = eval(monitor_metric_train_formatted, {}, {})

            monitor_metric_val_formatted = self.monitor_metric.replace('{', '{val_').format(**logs)
            new_model_monitor_val = eval(monitor_metric_val_formatted, {}, {})
            
        # Now we can save and clean models
        save_and_clean_models(
                model_save_dir=self.model_save_dir,
                segment_subject=self.segment_subject,
                traindata_id=self.traindata_id,
                architecture_id=self.architecture_id,
                trainparams_id=self.trainparams_id,
                monitor_metric_mode=self.monitor_metric_mode,
                new_model=self.model,
                new_model_monitor_train=new_model_monitor_train,
                new_model_monitor_val=new_model_monitor_val,
                new_model_epoch=epoch,
                save_format=self.save_format,
                save_best_only=self.save_best_only,
                save_min_accuracy=self.save_min_accuracy,
                save_min_accuracy_ignored_epoch=self.save_min_accuracy_ignored_epoch,        
                save_weights_only=self.save_weights_only,
                model_template_for_save=self.model_template_for_save,      
                verbose=self.verbose,
                only_report=self.only_report)
        
def save_and_clean_models(
        model_save_dir: Path,
        segment_subject: str,
        traindata_id: int, 
        architecture_id: int,
        trainparams_id: int,
        monitor_metric_mode: str,
        new_model = None,        
        new_model_monitor_train: Optional[float] = None,
        new_model_monitor_val: Optional[float] = None,
        new_model_epoch: Optional[int] = None,
        save_format: str = 'h5',
        save_best_only: bool = False,
        save_min_accuracy: float = 0.9,
        save_min_accuracy_ignored_epoch: int = 100,
        save_weights_only: bool = False,
        model_template_for_save = None, 
        verbose: bool = True,
        debug: bool = False,
        only_report: bool = False):
    """
    Save the new model if it is good enough... and cleanup existing models 
    if they are worse than the new or other existing models.
    
    Args
        model_save_dir (Path): dir containing the models
        segment_subject (str): segment subject 
        traindata_id (int): train data id
        architecture_id (int): model architecture id
        trainparams_id (int): id of the train params
        model_monitor_metric_mode (MetricMode): use 'min' if the monitored 
            metrics should be as low as possible, 'max' if a higher values 
            is better. 
        new_model (optional): the keras model object that will be saved
        new_model_monitor_train (float, optional): the monitored metric on the 
            train dataset
        new_model_monitor_val (float, optional): the monitored metric on the 
            validation dataset
        new_model_epoch (int, optional): the epoch in the training
        save_format (SaveFormat, optional): The format to save in: 
            * h5: keras format (= default)
            * tf: tensorflow savedmodel
        save_best_only (bool, optional): only keep the best model
        save_min_accuracy (float, optional): minimum accuracy to save the 
            model. Defaults to 0.9.
        save_min_accuracy_ignored_epoch (int, optional): at this epoch the 
            save_min_accuracy is ignored, to make sure there is a model saved, 
            even if the minimum accuracy is never reached. 
        save_weights_only (bool, optional): only save weights
        model_template_for_save (optional): if using multi-GPU training, pass
            the original model here to use this as template for saving        
        verbose (bool, optional): report the best model after save and cleanup
        debug (bool, optional): write debug logging
        only_report (bool, optional): only report which models would be cleaned up
    """
    # TODO: add option to specify minimum accuracy/iou score before saving to speed up, because saving takes quite some time! 
    # Check validaty of input
    monitor_metric_mode_values = ['min', 'max']
    if(monitor_metric_mode not in monitor_metric_mode_values):
        raise Exception(f"Invalid value for mode: {monitor_metric_mode}, should be one of {monitor_metric_mode_values}")
    save_format_values = ('h5', 'tf')
    if save_format not in save_format_values:
        raise Exception(f"Invalid value for save_format: {save_format}, should be one of {save_format_values}")
            
    # Get a list of all existing models
    model_info_list = get_models(
            model_dir=model_save_dir, 
            segment_subject=segment_subject,
            traindata_id=traindata_id,
            trainparams_id=trainparams_id)
    
    # If there is a new model passed as param, add it to the list
    new_model_path = None
    new_model_acc_combined = None
    if new_model is not None:
        
        if(new_model_monitor_train is None
           or new_model_monitor_val is None
           or new_model_epoch is None):
            raise Exception(f"If new_model is not None, new_model_monitor_... parameters cannot be None either???, new_model_monitor_train: {new_model_monitor_train}, new_model_monitor_val: {new_model_monitor_val}, new_model_epoch: {new_model_epoch}")
                    
        # Calculate combined accuracy
        new_model_monitor_combined = (new_model_monitor_train+new_model_monitor_val)/2
        
        # Build save filepath
        # Remark: accuracy values should always be as high as possible, so 
        # recalculate values if monitor_metric_mode is 'min'
        if monitor_metric_mode == 'max':
            new_model_acc_combined = new_model_monitor_combined
            new_model_acc_train = new_model_monitor_train
            new_model_acc_val = new_model_monitor_val
        else:
            new_model_acc_combined = 1-new_model_monitor_combined
            new_model_acc_train = 1-new_model_monitor_train
            new_model_acc_val = 1-new_model_monitor_val
        
        new_model_filename = format_model_filename(
                segment_subject=segment_subject,
                traindata_id=traindata_id,
                architecture_id=architecture_id,
                trainparams_id=trainparams_id,
                acc_combined=new_model_acc_combined,
                acc_train=new_model_acc_train, 
                acc_val=new_model_acc_val, 
                epoch=new_model_epoch,
                save_format=save_format)
        new_model_path = Path(model_save_dir) / new_model_filename
        
        # Append model to the retrieved models...
        model_info_list.append(
                {   'filepath': str(new_model_path),
                    'filename': new_model_filename,
                    'segment_subject': segment_subject,
                    'traindata_id': traindata_id,
                    'architecture_id': architecture_id,
                    'trainparams_id': trainparams_id,
                    'acc_combined': new_model_acc_combined,
                    'acc_train': new_model_acc_train,
                    'acc_val': new_model_acc_val,
                    'epoch': new_model_epoch,
                    'save_format': save_format 
                })
    
    # Loop through all existing models
    # Remark: the list is sorted descending before iterating it, this way new
    # modelss are saved bevore deleting the previous best one(s)
    model_info_df = pd.DataFrame(model_info_list)
    model_info_sorted_df = model_info_df.sort_values(by='acc_combined', ascending=False)
    for model_info in model_info_sorted_df.itertuples(index=False):
       
        # If only the best needs to be kept, check only on acc_combined...
        keep_model = True
        better_ones_df = None
        if save_best_only:
            better_ones_df = model_info_df[
                    (model_info_df.filepath != model_info.filepath) 
                     & (model_info_df.acc_combined >= model_info.acc_combined)]
            if len(better_ones_df) > 0:
                keep_model = False

        # If model is (relatively) ok, keep it
        if keep_model is True:
            logger.debug(f"KEEP {model_info.filename}")

            # If it is the new model that needs to be kept, keep it or save to disk
            if(new_model_path is not None and new_model_epoch is not None
               and only_report is not True
               and model_info.filepath == str(new_model_path)
               and not new_model_path.exists()):
                if(new_model_epoch > save_min_accuracy_ignored_epoch
                   or model_info.acc_combined > save_min_accuracy):
                    logger.debug('Save model start')
                    if save_weights_only:
                        if model_template_for_save is not None:
                            model_template_for_save.save_weights(str(new_model_path))
                        else:                    
                            new_model.save_weights(str(new_model_path))
                    else:
                        if model_template_for_save is not None:
                            model_template_for_save.save(str(new_model_path))
                        else:                    
                            new_model.save(str(new_model_path))
                    logger.debug('Save model ready')
                else:
                    print(f"New model is best model, but acc_combined < save_min_accuracy: {new_model_acc_combined} < {save_min_accuracy}")
        else:     
            # Bad model... can be removed (or not saved)
            if only_report is True:
                logger.debug(f"DELETE {model_info.filename}")
            elif Path(model_info.filepath).exists() is True:
                logger.debug(f"DELETE {model_info.filename}")
                if Path(model_info.filepath).is_dir() is True:
                    shutil.rmtree(model_info.filepath)
                else:
                    Path(model_info.filepath).unlink()
                
            if debug is True and better_ones_df is not None:
                print(f"Better one(s) found for{model_info.filename}:")
                for better_one in better_ones_df.itertuples(index=False):
                    print(f"  {better_one.filename}")

    if verbose is True or debug is True:
        best_model = get_best_model(
                model_dir=model_save_dir,
                segment_subject=segment_subject,
                traindata_id=traindata_id,
                architecture_id=architecture_id,
                trainparams_id=trainparams_id)
        if best_model is not None:
            logger.info(f"Current best model for {segment_subject}_{traindata_id}: acc_combined: {best_model['acc_combined']}, epoch: {best_model['epoch']}")

if __name__ == '__main__':
    raise Exception("Not implemented")




# Import the high level API
from orthoseg.load_images import load_images
from orthoseg.train import train
from orthoseg.train import _search_label_files
from orthoseg.predict import predict
from orthoseg.postprocess import postprocess


from pathlib import Path
import requests

def download_file(
        id: str, 
        destination: Path):

    URL = "https://docs.google.com/uc?export=download"

    session = requests.Session()

    response = session.get(URL, params = { 'id' : id }, stream = True)
    token = _get_confirm_token(response)

    if token:
        params = { 'id' : id, 'confirm' : token }
        response = session.get(URL, params = params, stream = True)

    destination.parent.mkdir(parents=True, exist_ok=True)
    _save_response_content(response, destination)    

def _get_confirm_token(response):
    for key, value in response.cookies.items():
        if key.startswith('download_warning'):
            return value

    return None

def _save_response_content(response, destination):
    CHUNK_SIZE = 32768

    with open(destination, "wb") as f:
        for chunk in response.iter_content(CHUNK_SIZE):
            if chunk: # filter out keep-alive new chunks
                f.write(chunk)



# -*- coding: utf-8 -*-
"""
Module with functions for post-processing prediction masks towards polygons.
"""

import logging
import math
from pathlib import Path
import shutil
from typing import List, Optional
import warnings

# Evade having many info warnings about self intersections from shapely
from geofileops import geofileops
from geofileops import geofile
from geofileops.util import geoseries_util as gfo_geoseries_util
import geopandas as gpd
import numpy as np
import rasterio as rio
import rasterio.features as rio_features
import rasterio.transform as rio_transform
import skimage.filters.rank
from skimage.morphology import rectangle
import shapely.geometry as sh_geom
import tensorflow as tf

from orthoseg.util import vector_util


logging.getLogger('shapely.geos').setLevel(logging.WARNING)

# -------------------------------------------------------------
# First define/init some general variables/constants
# -------------------------------------------------------------
# Get a logger...
logger = logging.getLogger(__name__)
# logger.setLevel(logging.DEBUG)

# -------------------------------------------------------------
# The real work
# -------------------------------------------------------------

# -------------------------------------------------------------
# Postprocess to use on all vector outputs
# -------------------------------------------------------------


def postprocess_predictions(
        input_path: Path,
        output_path: Path,
        dissolve: bool,
        dissolve_tiles_path: Optional[Path] = None,
        simplify_algorithm: Optional[geofileops.SimplifyAlgorithm] = None,
        simplify_tolerance: float = 1,
        simplify_lookahead: int = 8,
        nb_parallel: int = -1,
        force: bool = False) -> List[Path]:
    """
    Postprocesses the input prediction as specified. 
        
    Args
        input_path: path to the 'raw' prediction vector file.
        output_path: the base path where the output file(s) will be written to.
        dissolve (bool): True if a dissolve needs to be applied
        dissolve_tiles_path (PathLike, optional): Path to a geofile containing 
            the tiles to be used for the dissolve. Defaults to None.
        nb_parallel (int, optional): number of cpu's to use for postprocessing. 
            Use all cpu's if it is -1. Defaults to -1. 
        force: False to skip results that already exist, true to
               ignore existing results and overwrite them           
    """
    # Init
    if not input_path.exists():
        raise Exception(f"input_path does not exist: {input_path}")

    # The return value is the list of paths created
    output_paths = []

    # Because the geo operations will be applied sequentially if applicable, 
    # both the input path and output path will build on the result of the 
    # previous operation. 
    # Set the initial values to the ones passed in as parameters.  
    curr_input_path = input_path
    curr_output_path = output_path

    # Dissolve the predictions if needed
    if dissolve:
        curr_output_path = output_path.parent / f"{output_path.stem}_dissolve{output_path.suffix}"
        
        # If the dissolved file doesn't exist yet, go for it... 
        if not curr_output_path.exists():
            # If column classname present, group on it...
            layerinfo = geofile.get_layerinfo(input_path)
            if 'classname' in layerinfo.columns:
                groupby_columns = ['classname']
            else:
                groupby_columns = []
            columns = groupby_columns

            # Now we can dissolve
            geofileops.dissolve(
                    input_path=input_path,
                    tiles_path=dissolve_tiles_path,
                    output_path=curr_output_path,
                    groupby_columns=groupby_columns,
                    columns=columns,
                    explodecollections=True,
                    nb_parallel=nb_parallel,
                    force=force)

            # Add/recalculate columns with area and nbcoords
            geofile.add_column(
                    path=curr_output_path, 
                    name='area', type=geofile.DataType.REAL, expression='ST_Area(geom)',
                    force_update=True)
            geofile.add_column(
                    path=curr_output_path, 
                    name='nbcoords', type=geofile.DataType.INTEGER, expression='ST_NPoints(geom)',
                    force_update=True)

        # The curr_output_path becomes the new current input path 
        curr_input_path = curr_output_path
        output_paths.append(curr_output_path)

    # If a simplify algorithm is specified, simplify!
    if simplify_algorithm is not None:
        curr_input_path = curr_output_path
        curr_output_path = curr_output_path.parent / f"{curr_output_path.stem}_simpl{curr_output_path.suffix}"
        
        # If the simplified file doesn't exist yet, go for it... 
        if not curr_output_path.exists():
            # Simplify!
            geofileops.simplify(
                    input_path=curr_input_path,
                    output_path=curr_output_path,
                    algorithm=simplify_algorithm,
                    tolerance=simplify_tolerance,
                    lookahead=simplify_lookahead,
                    nb_parallel=nb_parallel)
            
            # Add/recalculate columns with area and nbcoords
            geofile.add_column(
                    path=curr_output_path, 
                    name='area', type=geofile.DataType.REAL, expression='ST_Area(geom)',
                    force_update=True)
            geofile.add_column(
                    path=curr_output_path, 
                    name='nbcoords', type=geofile.DataType.INTEGER, expression='ST_NPoints(geom)',
                    force_update=True)
            
        curr_input_path = curr_output_path
        output_paths.append(curr_output_path)
        
    return output_paths

def read_prediction_file(
        filepath: Path,
        border_pixels_to_ignore: int = 0) -> Optional[gpd.GeoDataFrame]:
    ext_lower = filepath.suffix.lower()
    if ext_lower == '.geojson':
        return geofile.read_file(filepath)
    elif ext_lower == '.tif':
        return polygonize_pred_from_file(filepath, border_pixels_to_ignore)
    else:
        raise Exception(f"Unsupported extension: {ext_lower}")

def to_binary_uint8(
        in_arr: np.ndarray, 
        thresshold_ok: int = 128) -> np.ndarray:

    # Check input parameters
    if in_arr.dtype != np.uint8:
        raise Exception("Input should be dtype = uint8, not: {in_arr.dtype}")
        
    # First copy to new numpy array, otherwise input array is changed
    out_arr = np.copy(in_arr)
    out_arr[out_arr >= thresshold_ok] = 255
    out_arr[out_arr < thresshold_ok] = 0
    
    return out_arr

def postprocess_for_evaluation(
        image_filepath: Path,
        image_crs: str,
        image_transform,
        image_pred_filepath: Path,
        image_pred_uint8_cleaned_bin: np.ndarray,
        class_id: int,
        class_name: str,
        nb_classes: int,
        output_dir: Path,
        output_suffix: str = None,
        input_image_dir: Optional[Path] = None,
        input_mask_dir: Optional[Path] = None,
        border_pixels_to_ignore: int = 0,
        force: bool = False):
    """
    This function postprocesses a prediction to make it easy to evaluate 
    visually if the result is OK by creating images of the different stages of 
    the prediction logic by creating the following output:
        - the input image
        - the mask image as digitized in the train files (if available)
        - the "raw" prediction image
        - the "raw" polygonized prediction, as an image
        - the simplified polygonized prediction, as an image

    The filenames start with a prefix:
        - if a mask is available, the % overlap between the result and the mask
        - if no mask is available, the % of pixels that is white
    
    Args
        
    """
    
    logger.debug(f"Start postprocess for {image_pred_filepath}")
    all_black = False
    try:      
        
        # If the image wasn't saved, it must have been all black
        if image_pred_filepath is None:
            all_black = True

        # Make sure the output dir exists...
        output_dir.mkdir(parents=True, exist_ok=True)
                
        # Determine the prefix to use for the output filenames
        pred_prefix_str = ''
        '''
        def jaccard_similarity(im1: np.ndarray, im2: np.ndarray):
            if im1.shape != im2.shape:
                message = f"Shape mismatch: input have different shape: im1: {im1.shape}, im2: {im2.shape}"
                logger.critical(message)
                raise ValueError(message)

            intersection = np.logical_and(im1, im2)
            union = np.logical_or(im1, im2)

            sum_union = float(union.sum())
            if sum_union == 0.0:
                # If 0 positive pixels in union: perfect prediction, so 1
                return 1
            else:
                sum_intersect = intersection.sum()
                return sum_intersect/sum_union
        '''

        # If there is a mask dir specified... use the groundtruth mask
        if input_mask_dir is not None and input_mask_dir.exists():
            # Read mask file and get all needed info from it...
            mask_filepath = Path(str(image_filepath).
                    replace(str(input_image_dir), str(input_mask_dir)))
                
            # Check if this file exists, if not, look for similar files
            if not mask_filepath.exists():
                files = list(mask_filepath.parent.glob(mask_filepath.stem + '*'))
                if len(files) == 1:
                    mask_filepath = files[0]
                else:
                    message = f"Error finding mask file with {mask_filepath.stem + '*'}: {len(files)} mask(s) found"
                    logger.error(message)
                    raise Exception(message)

            with rio.open(mask_filepath) as mask_ds:
                # Read pixels
                mask_arr = mask_ds.read(1)

            # Make the pixels at the borders of the mask black so they are 
            # ignored in the comparison
            if border_pixels_to_ignore and border_pixels_to_ignore > 0:
                mask_arr[0:border_pixels_to_ignore,:] = 0    # Left border
                mask_arr[-border_pixels_to_ignore:,:] = 0    # Right border
                mask_arr[:,0:border_pixels_to_ignore] = 0    # Top border
                mask_arr[:,-border_pixels_to_ignore:] = 0    # Bottom border
                
            # If there is more than 1 class, extract the seperate masks
            # per class with one-hot encoding
            if nb_classes > 1:
                mask_categorical_arr = tf.keras.utils.to_categorical(mask_arr, nb_classes, dtype=rio.uint8)
                mask_arr = (mask_categorical_arr[:,:,class_id]) * 255
                                
            #similarity = jaccard_similarity(mask_arr, image_pred)
            # Use accuracy as similarity... is more practical than jaccard
            similarity = np.array(np.equal(mask_arr, image_pred_uint8_cleaned_bin)
                                 ).sum() / image_pred_uint8_cleaned_bin.size
            pred_prefix_str = f"{similarity:0.3f}_"
            
            # Write mask
            mask_copy_dest_filepath = (output_dir / 
                    f"{pred_prefix_str}{image_filepath.stem}_{class_name}_mask.tif")
            #if not mask_copy_dest_filepath.exists():     
            with rio.open(mask_copy_dest_filepath, 'w', driver='GTiff', 
                    compress='lzw',
                    height=mask_arr.shape[1], width=mask_arr.shape[0], 
                    count=1, dtype=rio.uint8, 
                    crs=image_crs, transform=image_transform) as dst:
                dst.write(mask_arr, 1)

        else:
            # If all_black, no need to calculate again
            if all_black: 
                pct_black = 1
            else:
                # Calculate percentage black pixels
                pct_black = 1 - ((image_pred_uint8_cleaned_bin.sum()/250)
                                    /image_pred_uint8_cleaned_bin.size)
            
            # If the result after segmentation is all black, set all_black
            if pct_black == 1:
                # Force the prefix to be really high so it is clear they are entirely black
                pred_prefix_str = "1.001_"
                all_black = True
            else:
                pred_prefix_str = f"{pct_black:0.3f}_"

            # If there are few white pixels, don't save it,
            # because we are in evaluetion mode anyway...
            #if similarity >= 0.95:
            # continue
        
        # Copy the input image if it doesn't exist yet in output path
        output_basefilepath = output_dir / f"{pred_prefix_str}{image_filepath.stem}{output_suffix}"
        image_dest_filepath = Path(str(output_basefilepath) + image_filepath.suffix)
        if not image_dest_filepath.exists():
            shutil.copyfile(image_filepath, image_dest_filepath)

        # Rename the prediction file so it also contains the prefix,... 
        if image_pred_filepath is not None:
            image_dest_filepath = Path(f"{str(output_basefilepath)}_pred{image_pred_filepath.suffix}")
            if not image_dest_filepath.exists():
                shutil.move(str(image_pred_filepath), image_dest_filepath)

        # If all_black, we are ready now
        if all_black:
            logger.debug("All black prediction, no use proceding")
            return 
        
        # Write a cleaned-up version for evaluation as well 
        #polygonize_pred_for_evaluation(
        #        image_pred_uint8_bin=image_pred_uint8_cleaned_bin,
        #        image_crs=image_crs,
        #        image_transform=image_transform,
        #        output_basefilepath=output_basefilepath)
    
    except Exception as ex:
        message = f"Exception postprocessing prediction for {image_filepath}\n: file {image_pred_filepath}!!!"
        raise Exception(message) from ex

def polygonize_pred_for_evaluation(
        image_pred_uint8_bin,
        image_crs: str,
        image_transform,
        output_basefilepath: Path):

    # Polygonize result
    try:
        # Returns a list of tupples with (geometry, value)
        polygonized_records = list(rio_features.shapes(
                image_pred_uint8_bin, mask=image_pred_uint8_bin, transform=image_transform))
        
        # If nothing found, we can return
        if len(polygonized_records) == 0:
            logger.debug("This prediction didn't result in any polygons")
            return

        # Convert shapes to geopandas geodataframe 
        geoms = []
        for geom, _ in polygonized_records:
            geoms.append(sh_geom.shape(geom))   
        geoms_gdf = gpd.GeoDataFrame(geoms, columns=['geometry'])
        geoms_gdf.crs = image_crs

        image_shape = image_pred_uint8_bin.shape
        image_width = image_shape[0]
        image_height = image_shape[1]

        # For easier evaluation, write the cleaned version as raster
        # Write the standard cleaned output to file
        logger.debug("Save binary prediction")
        image_pred_cleaned_filepath = Path(f"{str(output_basefilepath)}_pred_bin.tif")
        with rio.open(image_pred_cleaned_filepath, 'w', driver='GTiff', 
                compress='lzw',
                height=image_height, width=image_width, 
                count=1, dtype=rio.uint8, 
                crs=image_crs, transform=image_transform) as dst:
            dst.write(image_pred_uint8_bin, 1)
        
        # If the input image contained a tranform, also create an image 
        # based on the simplified vectors
        if(image_transform[0] != 0 
                and len(geoms) > 0):
            # Simplify geoms
            geoms_simpl = []
            geoms_simpl_vis = []
            for geom in geoms:
                # The simplify of shapely uses the deuter-pecker algo
                # preserve_topology is slower bu makes sure no polygons are removed
                geom_simpl = geom.simplify(0.5, preserve_topology=True)
                if not geom_simpl.is_empty:
                    geoms_simpl.append(geom_simpl)
            
            # Write simplified wkt result to raster for comparing. 
            if len(geoms_simpl) > 0:
                # TODO: doesn't support multiple classes
                logger.debug('Before writing simpl rasterized file')
                image_pred_simpl_filepath = f"{str(output_basefilepath)}_pred_cleaned_simpl.tif"
                with rio.open(image_pred_simpl_filepath, 'w', driver='GTiff', compress='lzw',
                                height=image_height, width=image_width, 
                                count=1, dtype=rio.uint8, crs=image_crs, transform=image_transform) as dst:
                    # this is where we create a generator of geom, value pairs to use in rasterizing
                    logger.debug('Before rasterize')
                    burned = rio_features.rasterize(
                            shapes=geoms_simpl, 
                            out_shape=(image_height, image_width),
                            fill=0, default_value=255, dtype=rio.uint8,
                            transform=image_transform)
                    dst.write(burned, 1)
            
            # Write simplified wkt result to raster for comparing. Use the same
            if len(geoms_simpl_vis) > 0:
                # file profile as created before for writing the raw prediction result
                # TODO: doesn't support multiple classes
                logger.debug('Before writing simpl with visvangali algo rasterized file')
                image_pred_simpl_filepath = f"{str(output_basefilepath)}_pred_cleaned_simpl_vis.tif"
                with rio.open(image_pred_simpl_filepath, 'w', driver='GTiff', compress='lzw',
                              height=image_height, width=image_width, 
                              count=1, dtype=rio.uint8, crs=image_crs, transform=image_transform) as dst:
                    # this is where we create a generator of geom, value pairs to use in rasterizing
                    logger.debug('Before rasterize')
                    burned = rio_features.rasterize(
                            shapes=geoms_simpl_vis, 
                            out_shape=(image_height, image_width),
                            fill=0, default_value=255, dtype=rio.uint8,
                            transform=image_transform)
                    dst.write(burned, 1)

    except Exception as ex:
        message = f"Exception while polygonizing to file {output_basefilepath}!"
        raise Exception(message) from ex

def polygonize_pred_from_file(
        image_pred_filepath: Path,
        border_pixels_to_ignore: int = 0,
        save_to_file: bool = False) -> Optional[gpd.GeoDataFrame]:

    try:
        with rio.open(image_pred_filepath) as image_ds:
            # Read geo info
            image_crs = image_ds.profile['crs']
            image_transform = image_ds.transform
            
            # Read pixels and change from (channels, width, height) to 
            # (width, height, channels) and normalize to values between 0 and 1
            image_data = image_ds.read()
    
         # Create binary version
        #image_data = rio_plot.reshape_as_image(image_data)
        image_pred_uint8_bin = to_binary_uint8(image_data, 125)

        output_basefilepath = None
        if save_to_file is True:
            output_basefilepath = image_pred_filepath.parent / image_pred_filepath.stem
        result_gdf = polygonize_pred(
                image_pred_uint8_bin=image_pred_uint8_bin,
                image_crs=image_crs,
                image_transform=image_transform,
                output_basefilepath=output_basefilepath)

        if result_gdf is None:
            logger.warn(f"Prediction didn't result in any polygons: {image_pred_filepath}")
            
        return result_gdf

    except Exception as ex:
        raise Exception(f"Error in polygonize_pred_from_file on {image_pred_filepath}") from ex

def polygonize_pred_multiclass_to_file(
        image_pred_arr: np.ndarray,
        image_crs: str,
        image_transform,
        classes: list,
        output_vector_path: Path,
        min_probability: float = 0.5,
        prediction_cleanup_params: Optional[dict] = None,
        border_pixels_to_ignore: int = 0) -> bool:

    # Polygonize the result...
    result_gdf = polygonize_pred_multiclass(
            image_pred_uint8=image_pred_arr,
            image_crs=image_crs,
            image_transform=image_transform,
            classes=classes,
            min_probability=min_probability,
            prediction_cleanup_params=prediction_cleanup_params,
            border_pixels_to_ignore=border_pixels_to_ignore)

    # If there were polygons, save them...
    if result_gdf is not None:
        # TODO: review with new version of geopandas (> 0.10.2) if filtering  
        # this warning is still needed!
        # Evade pandas warnings: 
        #   "pandas.Int64Index is deprecated and will be removed from pandas 
        #   in a future version."
        with warnings.catch_warnings():
            warnings.filterwarnings(action='ignore', category=FutureWarning)
            geofile.to_file(result_gdf, output_vector_path, append=True, index=False)
        return True
    else:
        return False

def polygonize_pred_multiclass(
        image_pred_uint8: np.ndarray,
        image_crs: str,
        image_transform,
        classes: list,
        min_probability: float = 0.5,
        prediction_cleanup_params: Optional[dict] = None,
        border_pixels_to_ignore: int = 0) -> Optional[gpd.GeoDataFrame]:

    # Init
    result_gdf = None

    """
    for channel_id in range(0, nb_channels):
        image_pred_curr_arr = image_pred_arr[:,:,channel_id]

        # Clean prediction
        image_pred_uint8_cleaned_curr = clean_prediction(
                image_pred_arr=image_pred_curr_arr, 
                border_pixels_to_ignore=border_pixels_to_ignore)
    """

    # Reverse the one-hot decoding so each class has it's own number in the array,
    # but ignore prediction probability < min_probability
    image_pred_uint8[image_pred_uint8 < math.floor(255*min_probability)] = 0
    image_pred_decoded_arr = np.argmax(image_pred_uint8, axis=2).astype(np.uint8)
    
    # Make the pixels at the borders of the prediction black so they are ignored
    if border_pixels_to_ignore and border_pixels_to_ignore > 0:
        image_pred_decoded_arr[0:border_pixels_to_ignore,:] = 0    # Left border
        image_pred_decoded_arr[-border_pixels_to_ignore:,:] = 0    # Right border
        image_pred_decoded_arr[:,0:border_pixels_to_ignore] = 0    # Top border
        image_pred_decoded_arr[:,-border_pixels_to_ignore:] = 0    # Bottom border

    if prediction_cleanup_params is not None:
        # If fill_gaps_modal_size is asked... 
        if (
            "filter_background_modal_size" in prediction_cleanup_params 
            and prediction_cleanup_params["filter_background_modal_size"] is not None
            and prediction_cleanup_params["filter_background_modal_size"] > 0
        ):
            filter_background_modal_size = prediction_cleanup_params["filter_background_modal_size"]
            image_pred_decoded_modal_arr = skimage.filters.rank.modal(
                image_pred_decoded_arr, rectangle(
                    filter_background_modal_size, filter_background_modal_size
                )
            )
            np.copyto(
                image_pred_decoded_arr, 
                image_pred_decoded_modal_arr, 
                where=image_pred_decoded_arr==0
            )

    # Polygonize 
    result_gdf = polygonize_pred(
            image_pred_uint8_bin=image_pred_decoded_arr,
            image_crs=image_crs,
            image_transform=image_transform,
            classnames=classes)

    # Calculate the bounds of the image in projected coordinates
    image_shape = image_pred_decoded_arr.shape
    image_width = image_shape[0]
    image_height = image_shape[1]
    image_bounds = rio_transform.array_bounds(
            image_height, image_width, image_transform)
    x_pixsize = get_pixelsize_x(image_transform)
    y_pixsize = get_pixelsize_y(image_transform)
    border_bounds = (
        image_bounds[0]+border_pixels_to_ignore*x_pixsize,
        image_bounds[1]+border_pixels_to_ignore*y_pixsize,
        image_bounds[2]-border_pixels_to_ignore*x_pixsize,
        image_bounds[3]-border_pixels_to_ignore*y_pixsize,
    )
    
    if prediction_cleanup_params is not None:
        # If a simplify is asked... 
        if (
            "simplify_algorithm" in prediction_cleanup_params 
            and prediction_cleanup_params["simplify_algorithm"] is not None
        ):
            # Define the bounds of the image as linestring, so points on this 
            # border are preserved during the simplify
            border_polygon = sh_geom.box(*border_bounds)
            assert border_polygon.exterior is not None
            border_lines = sh_geom.LineString(border_polygon.exterior.coords)
            simplify_topological = prediction_cleanup_params["simplify_topological"] 
            if simplify_topological is None:
                simplify_topological = True if len(classes) > 2 else False
            if simplify_topological:
                result_gdf.geometry = gfo_geoseries_util.simplify_topo_ext(
                    geoseries=result_gdf.geometry, 
                    algorithm=prediction_cleanup_params['simplify_algorithm'],
                    tolerance=prediction_cleanup_params['simplify_tolerance'], 
                    lookahead=prediction_cleanup_params['simplify_lookahead'], 
                    keep_points_on=border_lines,
                )
            else:
                result_gdf.geometry = gfo_geoseries_util.simplify_ext(
                    geoseries=result_gdf.geometry, 
                    algorithm=prediction_cleanup_params['simplify_algorithm'],
                    tolerance=prediction_cleanup_params['simplify_tolerance'], 
                    lookahead=prediction_cleanup_params['simplify_lookahead'],
                    keep_points_on=border_lines
                )

            # Remove geom rows that became empty after simplify + explode
            result_gdf = result_gdf[~result_gdf.geometry.is_empty] 
            result_gdf = result_gdf[~result_gdf.geometry.isna()]  
            if len(result_gdf) == 0:
                return None
            #result_gdf.reset_index(drop=True, inplace=True)
            result_gdf = result_gdf.explode(ignore_index=True) # type: ignore
            #result_gdf.reset_index(drop=True, inplace=True)

    # Now we can calculate the "onborder" property
    result_gdf = vector_util.calc_onborder(result_gdf, border_bounds) # type: ignore

    # Add the area
    result_gdf['area'] = result_gdf.geometry.area

    return result_gdf

def polygonize_pred(
        image_pred_uint8_bin,
        image_crs: str,
        image_transform,
        classnames: Optional[List[str]] = None,
        output_basefilepath: Optional[Path] = None) -> Optional[gpd.GeoDataFrame]:

    # Polygonize result
    try:
        # Returns a list of tupples with (geometry, value)
        polygonized_records = list(rio_features.shapes(
                image_pred_uint8_bin, mask=image_pred_uint8_bin, transform=image_transform))

        # If nothing found, we can return
        if len(polygonized_records) == 0:
            return None

        # Convert shapes to geopandas geodataframe 
        data = [(sh_geom.shape(geom), int(value)) for geom, value in polygonized_records]
        result_gdf = gpd.GeoDataFrame(data, columns=["geometry", "value"], crs=image_crs)

        # Add the classname if provided and area
        if classnames is not None:
            result_gdf["classname"] = [
                classnames[value] for _, value in result_gdf["value"].T.iteritems()
            ]
            result_gdf = result_gdf.drop(columns=["value"])
        
        assert isinstance(result_gdf, gpd.GeoDataFrame)
        return result_gdf
            
    except Exception as ex:
        message = f"Exception while polygonizing to file {output_basefilepath}"
        raise Exception(message) from ex

def clean_and_save_prediction(
        image_image_filepath: Path,
        image_crs: str,
        image_transform: str,
        output_dir: Path,
        image_pred_arr: np.ndarray,
        classes: list,
        input_image_dir: Optional[Path] = None,
        input_mask_dir: Optional[Path] = None,
        border_pixels_to_ignore: int = 0,
        min_probability: float = 0.5,
        evaluate_mode: bool = False,
        force: bool = False) -> bool:

    # If nb. channels in prediction > 1, skip the first as it is the background
    image_pred_shape = image_pred_arr.shape
    nb_channels = image_pred_shape[2]
    if nb_channels > 1:
        channel_start = 1
    else:
        channel_start = 0

    for channel_id in range(channel_start, nb_channels):
        image_pred_curr_arr = image_pred_arr[:,:,channel_id]

        # Clean prediction
        image_pred_uint8_cleaned_curr = clean_prediction(
                image_pred_arr=image_pred_curr_arr, 
                border_pixels_to_ignore=border_pixels_to_ignore)
        
        # If the cleaned result contains useful values or in evaluate mode... save
        if(min_probability == 0 
           or np.any(image_pred_uint8_cleaned_curr >= math.floor(min_probability*255))
           or evaluate_mode is True):

            # Find the class name in the classes list
            class_name = None
            for class_id, (classname) in enumerate(classes):
                if class_id == channel_id:
                    class_name = classname
                    break
            if class_name is None:
                raise Exception(f"No classname found for channel_id {channel_id}")
        
            # Now save prediction
            output_suffix=f"_{class_name}"
            image_pred_filepath = save_prediction_uint8(
                    image_filepath=image_image_filepath,
                    image_pred_uint8_cleaned=image_pred_uint8_cleaned_curr,
                    image_crs=image_crs,
                    image_transform=image_transform,
                    output_dir=output_dir,
                    output_suffix=output_suffix,
                    force=force)

            # Postprocess for evaluation
            if evaluate_mode is True:
                # Create binary version and postprocess
                image_pred_uint8_cleaned_bin = to_binary_uint8(image_pred_uint8_cleaned_curr, 125)                       
                postprocess_for_evaluation(
                        image_filepath=image_image_filepath,
                        image_crs=image_crs,
                        image_transform=image_transform,
                        image_pred_filepath=image_pred_filepath,
                        image_pred_uint8_cleaned_bin=image_pred_uint8_cleaned_bin,
                        output_dir=output_dir,
                        output_suffix=output_suffix,
                        input_image_dir=input_image_dir,
                        input_mask_dir=input_mask_dir,
                        class_id=channel_id,
                        class_name=class_name,
                        nb_classes=nb_channels,
                        border_pixels_to_ignore=border_pixels_to_ignore,
                        force=force)

    return True

def clean_prediction(
        image_pred_arr: np.ndarray,
        border_pixels_to_ignore: int = 0,
        output_color_depth: str = 'binary') -> np.ndarray:
    """
    Cleans a prediction result and returns a cleaned, uint8 array.
    
    Args:
        image_pred_arr (np.array): The prediction as returned by keras.
        border_pixels_to_ignore (int, optional): Border pixels to ignore. Defaults to 0.
        output_color_depth (str, optional): Color depth desired. Defaults to '2'.
            * binary: 0 or 255
            * full: 256 different values
    
    Returns:
        np.array: The cleaned result.
    """

    # Input should be float32
    if image_pred_arr.dtype not in [np.float32, np.uint8]:
        raise Exception(f"image prediction is in an unsupported type: {image_pred_arr.dtype}") 
    if output_color_depth not in ['binary', 'full']:
        raise Exception(f"Unsupported output_color_depth: {output_color_depth}")

    # Reshape from 3 to 2 dims if necessary (width, height, nb_channels).
    # Check the number of channels of the output prediction
    image_pred_shape = image_pred_arr.shape
    if len(image_pred_shape) > 2:
        n_channels = image_pred_shape[2]
        if n_channels > 1:
            raise Exception("Invalid input, should be one channel!")
        # Reshape array from 3 dims (width, height, nb_channels) to 2.
        image_pred_uint8 = np.reshape(image_pred_arr, (image_pred_shape[0], image_pred_shape[1]))   

    # Convert to uint8 if necessary
    if image_pred_arr.dtype == np.float32:
        image_pred_uint8 = np.array((image_pred_arr * 255), dtype=np.uint8)
    else:
        image_pred_uint8 = image_pred_arr

    # Convert to binary if needed
    if output_color_depth == 'binary':
        image_pred_uint8[image_pred_uint8 >= 127] = 255
        image_pred_uint8[image_pred_uint8 < 127] = 0
    
    # Make the pixels at the borders of the prediction black so they are ignored
    image_pred_uint8_cropped = image_pred_uint8
    if border_pixels_to_ignore and border_pixels_to_ignore > 0:
        image_pred_uint8_cropped[0:border_pixels_to_ignore,:] = 0    # Left border
        image_pred_uint8_cropped[-border_pixels_to_ignore:,:] = 0    # Right border
        image_pred_uint8_cropped[:,0:border_pixels_to_ignore] = 0    # Top border
        image_pred_uint8_cropped[:,-border_pixels_to_ignore:] = 0    # Bottom border
    
    return image_pred_uint8_cropped

def save_prediction_uint8(
        image_filepath: Path,
        image_pred_uint8_cleaned: np.ndarray,
        image_crs: str,
        image_transform: str,
        output_dir: Path,
        output_suffix: str = '',
        border_pixels_to_ignore: int = None,
        force: bool = False) -> Path:

    ##### Init #####
    # If no decent transform metadata, stop!
    if image_transform is None or image_transform[0] == 0:
        message = f"No transform found for {image_filepath}: {image_transform}"
        logger.error(message)
        raise Exception(message)

    # Make sure the output dir exists...
    if not output_dir.exists():
        output_dir.mkdir()
    
    # Write prediction to file
    output_filepath = output_dir / f"{image_filepath.stem}{output_suffix}_pred.tif"
    logger.debug("Save +- original prediction")
    image_shape = image_pred_uint8_cleaned.shape
    image_width = image_shape[0]
    image_height = image_shape[1]
    with rio.open(str(output_filepath), 'w', driver='GTiff', tiled='no',
                  compress='lzw', predictor=2, num_threads=4,
                  height=image_height, width=image_width, 
                  count=1, dtype=rio.uint8, crs=image_crs, transform=image_transform) as dst:
        dst.write(image_pred_uint8_cleaned, 1)
    
    return output_filepath

#-------------------------------------------------------------
# Helpers for working with Affine objects...                    
#-------------------------------------------------------------

def get_pixelsize_x(transform):
    return transform[0]
    
def get_pixelsize_y(transform):
    return -transform[4]

#-------------------------------------------------------------
# If the script is ran directly...
#-------------------------------------------------------------

if __name__ == '__main__':
    raise Exception("Not implemented")
